# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Iterator, List, Optional, Tuple

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3 import framework
from volatility3.framework import (
    constants,
    deprecation,
    exceptions,
    interfaces,
    renderers,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.linux import network

vollog = logging.getLogger(__name__)


@dataclass
class Proto:
    name: str
    hooks: Tuple[str] = field(default_factory=tuple)


PROTO_NOT_IMPLEMENTED = Proto(name="UNSPEC")

NF_INET_HOOKS = ("PRE_ROUTING", "LOCAL_IN", "FORWARD", "LOCAL_OUT", "POST_ROUTING")
NF_DEC_HOOKS = (
    "PRE_ROUTING",
    "LOCAL_IN",
    "FORWARD",
    "LOCAL_OUT",
    "POST_ROUTING",
    "HELLO",
    "ROUTE",
)
NF_ARP_HOOKS = ("IN", "OUT", "FORWARD")
NF_NETDEV_HOOKS = ("INGRESS", "EGRESS")
LARGEST_HOOK_NUMBER = max(
    len(NF_INET_HOOKS), len(NF_DEC_HOOKS), len(NF_ARP_HOOKS), len(NF_NETDEV_HOOKS)
)


class AbstractNetfilter(ABC):
    """Netfilter Abstract Base Classes handling details across various
    Netfilter implementations, including constants, helpers, and common
    routines.
    """

    PROTO_HOOKS = (
        PROTO_NOT_IMPLEMENTED,  # NFPROTO_UNSPEC
        Proto(name="INET", hooks=NF_INET_HOOKS),  # From kernels 3.14
        Proto(name="IPV4", hooks=NF_INET_HOOKS),
        Proto(name="ARP", hooks=NF_ARP_HOOKS),
        PROTO_NOT_IMPLEMENTED,
        Proto(name="NETDEV", hooks=NF_NETDEV_HOOKS),
        PROTO_NOT_IMPLEMENTED,
        Proto(name="BRIDGE", hooks=NF_INET_HOOKS),
        PROTO_NOT_IMPLEMENTED,
        PROTO_NOT_IMPLEMENTED,
        Proto(name="IPV6", hooks=NF_INET_HOOKS),
        PROTO_NOT_IMPLEMENTED,
        Proto(name="DECNET", hooks=NF_DEC_HOOKS),  # Removed in kernel 6.1
    )
    NF_MAX_HOOKS = LARGEST_HOOK_NUMBER + 1

    def __init__(
        self, context: interfaces.context.ContextInterface, kernel_module_name: str
    ):
        self._context = context
        self.vmlinux = context.modules[kernel_module_name]
        self.layer_name = self.vmlinux.layer_name

        # Set data sizes
        self.ptr_size = self.vmlinux.get_type("pointer").size
        self.list_head_size = self.vmlinux.get_type("list_head").size

        linuxutils_modulegatherers_required_version = (
            Netfilter._required_linuxutils_gatherers_version
        )
        linuxutils_modulegatherers_current_version = (
            linux_utilities_modules.ModuleGatherers.version
        )
        if not requirements.VersionRequirement.matches_required(
            linuxutils_modulegatherers_required_version,
            linuxutils_modulegatherers_current_version,
        ):
            raise exceptions.PluginRequirementException(
                f"linux_utilities_modules.ModuleGatherer version not suitable: required {linuxutils_modulegatherers_required_version} found {linuxutils_modulegatherers_current_version}"
            )

        linux_net_required_version = Netfilter._required_linuxnet_version
        linux_net_current_version = network.NetSymbols.version
        if not requirements.VersionRequirement.matches_required(
            linux_net_required_version, linux_net_current_version
        ):
            raise exceptions.PluginRequirementException(
                f"symbols.linux.net.NetSymbols version not suitable: required {linux_net_required_version} found {linux_net_current_version}"
            )

        linux_utilities_modules_required_version = (
            Netfilter._required_linux_utilities_modules_version
        )
        linux_utilities_modules_current_version = (
            linux_utilities_modules.Modules.version
        )
        if not requirements.VersionRequirement.matches_required(
            linux_utilities_modules_required_version,
            linux_utilities_modules_current_version,
        ):
            raise exceptions.PluginRequirementException(
                f"linux_utilities_modules.Modules version not suitable: required {linux_utilities_modules_required_version} found {linux_utilities_modules_current_version}"
            )

        symbol_table = context.symbol_space[self.vmlinux.symbol_table_name]
        network.NetSymbols.apply(symbol_table)

        self.handlers = linux_utilities_modules.Modules.run_modules_scanners(
            context=context,
            kernel_module_name=kernel_module_name,
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )

    @classmethod
    def run_all(
        cls, context: interfaces.context.ContextInterface, kernel_module_name: str
    ) -> Iterator[Tuple[int, str, str, int, int, str, bool]]:
        """It calls each subclass symtab_checks() to test the required
        conditions to that specific kernel implementation.

        Args:
            context: The volatility3 context on which to operate
            kernel_module_name: The name of the table containing the kernel symbols

        Yields:
            The kmsg records. Same as _run()
        """
        vmlinux = context.modules[kernel_module_name]

        implementation_inst = None  # type: ignore
        for subclass in framework.class_subclasses(cls):
            if not subclass.symtab_checks(vmlinux=vmlinux):
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    "Netfilter implementation '%s' doesn't match this memory dump",
                    subclass.__name__,
                )
                continue

            vollog.log(
                constants.LOGLEVEL_VVVV,
                "Netfilter implementation '%s' matches!",
                subclass.__name__,
            )
            implementation_inst = subclass(
                context=context, kernel_module_name=kernel_module_name
            )
            # More than one class could be executed for an specific kernel version
            # For instance: Netfilter Ingress hooks
            yield from implementation_inst._run()

        if implementation_inst is None:
            vollog.error("Unsupported Netfilter kernel implementation")

    def _run(self) -> Iterator[Tuple[int, str, str, int, int, str, bool]]:
        """Iterates over namespaces and protocols, executing various callbacks that
        allow  customization of the code to the specific data structure used in a
        particular kernel implementation

            get_hooks_container(net, proto_name, hook_name)
                It returns the data structure used in a specific kernel implementation
                to store the hooks for a respective namespace and protocol, basically:
                    For Ingress hooks:
                        network_namespace[] -> net_device[] -> nf_hooks_ingress[]
                    For egress hooks:
                        network_namespace[] -> net_device[] -> nf_hooks_egress[]
                    For all the other Netfilter hooks:
                        <= 4.2.8
                            nf_hooks[]
                        >= 4.3
                            network_namespace[] -> nf.hooks[]

            get_hook_ops(hook_container, proto_idx, hook_idx)
                Give the 'hook_container' got in get_hooks_container(), it
                returns an iterable of 'nf_hook_ops' elements for a respective protocol
                and hook type.

        Returns:
            netns [int]: Network namespace id
            proto_name [str]: Protocol name
            hook_name [str]: Hook name
            priority [int]: Priority
            hook_ops_hook [int]: Hook address
            module_name [str]: Linux kernel module name
            hooked [bool]: "True" if the network stack has been hijacked
        """
        for netns, net in self.get_net_namespaces():
            for proto_idx, proto_name, hook_idx, hook_name in self._proto_hook_loop():
                hooks_container = self.get_hooks_container(net, proto_name, hook_name)

                for hook_container in hooks_container:
                    for hook_ops in self.get_hook_ops(
                        hook_container, proto_idx, hook_idx
                    ):
                        if not hook_ops:
                            continue

                        priority = int(hook_ops.priority)
                        hook_ops_hook = hook_ops.hook
                        module_info, symbol_name = (
                            linux_utilities_modules.Modules.module_lookup_by_address(
                                self._context,
                                self.vmlinux.name,
                                self.handlers,
                                hook_ops_hook,
                            )
                        )
                        hooked = module_info is None

                        yield (
                            netns,
                            proto_name,
                            hook_name,
                            priority,
                            hook_ops_hook,
                            module_info,
                            symbol_name,
                            hooked,
                        )

    @classmethod
    @abstractmethod
    def symtab_checks(cls, vmlinux: interfaces.context.ModuleInterface) -> bool:
        """This method on each sublasss will be called to evaluate if the kernel
        being analyzed fulfill the type & symbols requirements for the implementation.
        The first class returning True will be instantiated and called via the
        run() method.

        Returns:
            bool: True if the kernel being analyzed fulfill the class requirements.
        """

    def _proto_hook_loop(self) -> Iterator[Tuple[int, str, int, str]]:
        """Flattens the protocol families and hooks"""
        for proto_idx, proto in enumerate(AbstractNetfilter.PROTO_HOOKS):
            if proto == PROTO_NOT_IMPLEMENTED:
                continue
            if proto.name not in self.subscribed_protocols():
                # This protocol is not managed in this object
                continue
            for hook_idx, hook_name in enumerate(proto.hooks):
                yield proto_idx, proto.name, hook_idx, hook_name

    def build_nf_hook_ops_array(
        self, nf_hook_entries
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Function helper to build the nf_hook_ops array when it is not part of the
        struct 'nf_hook_entries' definition.

        nf_hook_ops was stored adjacent in memory to the nf_hook_entry array, in the
        new struct 'nf_hook_entries'. However, this 'nf_hooks_ops' array 'orig_ops' is
        not part of the 'nf_hook_entries' struct. So, we need to calculate the offset.

            struct nf_hook_entries {
                u16                         num_hook_entries; /* plus padding */
                struct nf_hook_entry        hooks[];
                //const struct nf_hook_ops *orig_ops[];
            }
        """
        nf_hook_entry_size = self.vmlinux.get_type("nf_hook_entry").size

        try:
            num_hook_entries = nf_hook_entries.num_hook_entries
        except exceptions.InvalidAddressException:
            return None

        orig_ops_addr = (
            nf_hook_entries.hooks.vol.offset + nf_hook_entry_size * num_hook_entries
        )

        if not self.vmlinux._context.layers[self.vmlinux.layer_name].is_valid(
            orig_ops_addr
        ):
            return None

        orig_ops = self._context.object(
            object_type=self.get_symbol_fullname("array"),
            offset=orig_ops_addr,
            subtype=self.vmlinux.get_type("pointer"),
            layer_name=self.layer_name,
            count=num_hook_entries,
        )

        return orig_ops

    def subscribed_protocols(self) -> Tuple[str]:
        """Allows to select which PROTO_HOOKS protocols will be processed by the
        Netfiler subclass.
        """

        # Most implementation handlers respond to these protocols, except for
        # the ingress hook, which specifically handles the 'NETDEV' protocol.
        # However, there is no corresponding Netfilter hook implementation for
        # the INET protocol in the kernel. AFAIU, this is used as
        # 'NFPROTO_INET = NFPROTO_IPV4 || NFPROTO_IPV6'
        # in other parts of the kernel source code.
        return ("IPV4", "ARP", "BRIDGE", "IPV6", "DECNET")

    @deprecation.method_being_removed(
        removal_date="2026-03-25",
        message="Callers to this method should adapt `linux_utilities_modules.Modules.run_module_scanners`",
    )
    def get_module_name_for_address(self, addr) -> str:
        """Helper to obtain the module and symbol name in the format needed for the
        output of this plugin.
        """
        module_name, symbol_name = (
            linux_utilities_modules.Modules.lookup_module_address(
                self._context, self.vmlinux.name, self.handlers, addr
            )
        )

        if module_name == "UNKNOWN":
            module_name = None

        if symbol_name != "N/A":
            module_name = f"[{symbol_name}]"

        return module_name

    def get_net_namespaces(self):
        """Common function to retrieve the different namespaces.
        From 4.3 on, all the implementations use network namespaces.
        """
        nethead = self.vmlinux.object_from_symbol("net_namespace_list")
        symbol_net_name = self.get_symbol_fullname("net")
        for net in nethead.to_list(symbol_net_name, "list"):
            net_ns_id = net.ns.inum
            yield net_ns_id, net

    def get_hooks_container(self, net, proto_name, hook_name):
        """Returns the data structure used in a specific kernel implementation to store
        the hooks for a respective namespace and protocol.

        Except for kernels < 4.3, all the implementations use network namespaces.
        Also the data structure which contains the hooks, even though it changes its
        implementation and/or data type, it is always in this location.
        """
        yield net.nf.hooks

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        """Given the hook_container obtained from get_hooks_container(), it
        returns an iterable of 'nf_hook_ops' elements for a corresponding protocol
        and hook type.

        This is the most variable/unstable part of all Netfilter hook designs, it
        changes almost in every single implementation.
        """
        raise NotImplementedError("You must implement this method")

    def get_symbol_fullname(self, symbol_basename: str) -> str:
        """Given a short symbol or type name, it returns its full name"""
        return self.vmlinux.symbol_table_name + constants.BANG + symbol_basename

    @staticmethod
    def get_member_type(
        vol_type: interfaces.objects.Template, member_name: str
    ) -> List[str]:
        """Returns a list of types/subtypes belonging to the given type member.

        Args:
            vol_type (interfaces.objects.Template): A vol3 type object
            member_name (str): The member name

        Returns:
            list: A list of types/subtypes
        """
        _size, vol_obj = vol_type.vol.members[member_name]
        type_name = vol_obj.type_name
        type_basename = type_name.split(constants.BANG)[1]
        member_type = [type_basename]
        cur_type = vol_obj
        while hasattr(cur_type, "subtype"):
            subtype_name = cur_type.subtype.type_name
            subtype_basename = subtype_name.split(constants.BANG)[1]
            member_type.append(subtype_basename)
            cur_type = cur_type.subtype

        return member_type


class NetfilterImp_to_4_3(AbstractNetfilter):
    """At this point, Netfilter hooks were implemented as a linked list of struct
    'nf_hook_ops' type. One linked list per protocol per hook type.
    It was like that until 4.2.8.

        struct list_head nf_hooks[NFPROTO_NUMPROTO][NF_MAX_HOOKS];
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        return vmlinux.has_symbol("nf_hooks")

    def get_net_namespaces(self):
        # In kernels <= 4.2.8 netfilter hooks are not implemented per namespaces
        netns, net = renderers.NotAvailableValue(), renderers.NotAvailableValue()
        yield netns, net

    def get_hooks_container(self, net, proto_name, hook_name):
        nf_hooks = self.vmlinux.object_from_symbol("nf_hooks")
        if not nf_hooks:
            return

        yield nf_hooks

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        list_head = hook_container[proto_idx][hook_idx]
        nf_hooks_ops_name = self.get_symbol_fullname("nf_hook_ops")
        return list_head.to_list(nf_hooks_ops_name, "list")


class NetfilterImp_4_3_to_4_9(AbstractNetfilter):
    """Netfilter hooks were added to network namespaces in 4.3.
    It is still implemented as a linked list of 'struct nf_hook_ops' type but inside a
    network namespace. One linked list per protocol per hook type.

        struct net { ... struct netns_nf nf; ... }
        struct netns_nf { ...
            struct list_head hooks[NFPROTO_NUMPROTO][NF_MAX_HOOKS]; ... }
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("netns_nf")
            and vmlinux.get_type("netns_nf").has_member("hooks")
            and cls.get_member_type(vmlinux.get_type("netns_nf"), "hooks")
            == ["array", "array", "list_head"]
        )

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        list_head = hook_container[proto_idx][hook_idx]
        nf_hooks_ops_name = self.get_symbol_fullname("nf_hook_ops")
        return list_head.to_list(nf_hooks_ops_name, "list")


class NetfilterImp_4_9_to_4_14(AbstractNetfilter):
    """In this range of kernel versions, the doubly-linked lists of netfilter hooks were
    replaced by an array of arrays of 'nf_hook_entry' pointers in a singly-linked lists.
        struct net { ... struct netns_nf nf; ... }
        struct netns_nf { ..
            struct nf_hook_entry __rcu *hooks[NFPROTO_NUMPROTO][NF_MAX_HOOKS]; ... }

    Also in v4.10 the struct nf_hook_entry changed, a hook function pointer was added to
    it. However, for simplicity of this design, we will still take the hook address from
    the 'nf_hook_ops'. As per v5.0-rc2, the hook address is duplicated in both sides.
    - v4.9:
        struct nf_hook_entry {
            struct nf_hook_entry      *next;
            struct nf_hook_ops        ops;
            const struct nf_hook_ops  *orig_ops; };
    - v4.10:
        struct nf_hook_entry {
            struct nf_hook_entry      *next;
            nf_hookfn                 *hook;
            void                      *priv;
            const struct nf_hook_ops  *orig_ops; };
    (*) Even though the hook address is in the struct 'nf_hook_entry', we use the
    original 'nf_hook_ops' hook address value, the one which was filled by the user, to
    make it uniform to all the implementations.
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        hooks_type = ["array", "array", "pointer", "nf_hook_entry"]
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("netns_nf")
            and vmlinux.get_type("netns_nf").has_member("hooks")
            and cls.get_member_type(vmlinux.get_type("netns_nf"), "hooks") == hooks_type
        )

    def _get_hook_ops(self, hook_container, proto_idx, hook_idx):
        list_head = hook_container[proto_idx][hook_idx]
        nf_hooks_ops_name = self.get_symbol_fullname("nf_hook_ops")
        return list_head.to_list(nf_hooks_ops_name, "list")

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        nf_hook_entry_list = hook_container[proto_idx][hook_idx]
        while nf_hook_entry_list:
            yield nf_hook_entry_list.orig_ops
            nf_hook_entry_list = nf_hook_entry_list.next


class NetfilterImp_4_14_to_4_16(AbstractNetfilter):
    """'nf_hook_ops' was removed from struct 'nf_hook_entry'. Instead, it was stored
    adjacent in memory to the 'nf_hook_entry' array, in the new struct 'nf_hook_entries'
    However, 'orig_ops' is not part of the 'nf_hook_entries' struct definition. So, we
    have to craft it by hand.

        struct net { ... struct netns_nf nf; ... }
        struct netns_nf {
            struct nf_hook_entries *hooks[NFPROTO_NUMPROTO][NF_MAX_HOOKS]; ... }
        struct nf_hook_entries {
            u16                         num_hook_entries; /* plus padding */
            struct nf_hook_entry        hooks[];
            //const struct nf_hook_ops *orig_ops[]; }
        struct nf_hook_entry {
            nf_hookfn   *hook;
            void        *priv; }

    (*) Even though the hook address is in the struct 'nf_hook_entry', we use the
    original 'nf_hook_ops' hook address value, the one which was filled by the user, to
    make it uniform to all the implementations.
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        hooks_type = ["array", "array", "pointer", "nf_hook_entries"]
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("netns_nf")
            and vmlinux.get_type("netns_nf").has_member("hooks")
            and cls.get_member_type(vmlinux.get_type("netns_nf"), "hooks") == hooks_type
        )

    def get_nf_hook_entries(self, nf_hooks_addr, proto_idx, hook_idx):
        """This allows to support different hook array implementations from this version
        on. For instance, in kernels >= 4.16 this multi-dimensional array is split in
        one-dimensional array of pointers to 'nf_hooks_entries' per each protocol."""
        return nf_hooks_addr[proto_idx][hook_idx]

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        nf_hook_entries = self.get_nf_hook_entries(hook_container, proto_idx, hook_idx)
        if not nf_hook_entries:
            return

        nf_hook_ops_name = self.get_symbol_fullname("nf_hook_ops")
        nf_hook_ops_ptr_arr = self.build_nf_hook_ops_array(nf_hook_entries)
        if not nf_hook_ops_ptr_arr:
            return

        for nf_hook_ops_ptr in nf_hook_ops_ptr_arr:
            nf_hook_ops = nf_hook_ops_ptr.dereference().cast(nf_hook_ops_name)
            yield nf_hook_ops


class NetfilterImp_4_16_to_latest(NetfilterImp_4_14_to_4_16):
    """The multidimensional array of nf_hook_entries was split in a one-dimensional
    array per each protocol.

        struct net {
            struct netns_nf nf; ... }
        struct  netns_nf  {
            struct nf_hook_entries * hooks_ipv4[NF_INET_NUMHOOKS];
            struct nf_hook_entries * hooks_ipv6[NF_INET_NUMHOOKS];
            struct nf_hook_entries * hooks_arp[NF_ARP_NUMHOOKS];
            struct nf_hook_entries * hooks_bridge[NF_INET_NUMHOOKS];
            struct nf_hook_entries * hooks_decnet[NF_DN_NUMHOOKS]; ... }
        struct nf_hook_entries {
                u16 num_hook_entries; /* plus padding */
                struct nf_hook_entry hooks[];
                //const struct nf_hook_ops *orig_ops[]; }
        struct nf_hook_entry {
            nf_hookfn   *hook;
                void *priv; }

    (*) Even though the hook address is in the struct nf_hook_entry, we use the original
    nf_hook_ops hook address value, the one which was filled by the user, to make it
    uniform to all the implementations.
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("netns_nf")
            and vmlinux.get_type("netns_nf").has_member("hooks_ipv4")
        )

    def get_hooks_container(self, net, proto_name, hook_name):
        try:
            if proto_name == "IPV4":
                net_nf_hooks = net.nf.hooks_ipv4
            elif proto_name == "ARP":
                net_nf_hooks = net.nf.hooks_arp
            elif proto_name == "BRIDGE":
                net_nf_hooks = net.nf.hooks_bridge
            elif proto_name == "IPV6":
                net_nf_hooks = net.nf.hooks_ipv6
            elif proto_name == "DECNET":
                net_nf_hooks = net.nf.hooks_decnet
            else:
                return

            yield net_nf_hooks

        except AttributeError:
            # Protocol family disabled at kernel compilation
            #  CONFIG_NETFILTER_FAMILY_ARP=n ||
            #  CONFIG_NETFILTER_FAMILY_BRIDGE=n ||
            #  CONFIG_DECNET=n
            pass

    def _get_nf_hook_entries_ptr(self, nf_hooks_addr, proto_idx, hook_idx):
        nf_hook_entries_ptr = nf_hooks_addr[hook_idx]
        return nf_hook_entries_ptr

    def get_nf_hook_entries(self, nf_hooks_addr, proto_idx, hook_idx):
        return nf_hooks_addr[hook_idx]


class AbstractNetfilterNetDev(AbstractNetfilter):
    """Base class to handle the Netfilter NetDev hooks.
    It won't be executed. It has some common functions to all Netfilter NetDev hook
    implementations.

    Netfilter NetDev hooks are set per network device which belongs to a network
    namespace.
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        return False

    def subscribed_protocols(self):
        return ("NETDEV",)

    def get_hooks_container(self, net, proto_name, hook_name):
        net_device_type = self.vmlinux.get_type("net_device")
        net_device_name = self.get_symbol_fullname("net_device")
        for net_device in net.dev_base_head.to_list(net_device_name, "dev_list"):
            if hook_name == "INGRESS":
                if net_device_type.has_member("nf_hooks_ingress"):
                    # CONFIG_NETFILTER_INGRESS=y
                    yield net_device.nf_hooks_ingress

            elif hook_name == "EGRESS":
                if net_device_type.has_member("nf_hooks_egress"):
                    # CONFIG_NETFILTER_EGRESS=y
                    yield net_device.nf_hooks_egress


class NetfilterNetDevImp_4_2_to_4_9(AbstractNetfilterNetDev):
    """This is the first version of Netfilter Ingress hooks which was implemented using
    a doubly-linked list of 'nf_hook_ops'.
        struct list_head nf_hooks_ingress;
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        hooks_type = ["list_head"]
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("net_device")
            and vmlinux.get_type("net_device").has_member("nf_hooks_ingress")
            and cls.get_member_type(vmlinux.get_type("net_device"), "nf_hooks_ingress")
            == hooks_type
        )

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        nf_hooks_ingress = hook_container
        nf_hook_ops_name = self.get_symbol_fullname("nf_hook_ops")
        return nf_hooks_ingress.to_list(nf_hook_ops_name, "list")


class NetfilterNetDevImp_4_9_to_4_14(AbstractNetfilterNetDev):
    """In 4.9 it was changed to a simple singly-linked list.
    struct nf_hook_entry * nf_hooks_ingress;
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        hooks_type = ["pointer", "nf_hook_entry"]
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("net_device")
            and vmlinux.get_type("net_device").has_member("nf_hooks_ingress")
            and cls.get_member_type(vmlinux.get_type("net_device"), "nf_hooks_ingress")
            == hooks_type
        )

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        nf_hooks_ingress_ptr = hook_container
        if not nf_hooks_ingress_ptr:
            return

        while nf_hooks_ingress_ptr:
            nf_hook_entry = nf_hooks_ingress_ptr.dereference()
            orig_ops = nf_hook_entry.orig_ops.dereference()
            yield orig_ops
            nf_hooks_ingress_ptr = nf_hooks_ingress_ptr.next


class NetfilterNetDevImp_4_14_to_latest(AbstractNetfilterNetDev):
    """In 4.14 the hook list was converted to an array of pointers inside the struct
    'nf_hook_entries':
    struct nf_hook_entries * nf_hooks_ingress;
    struct nf_hook_entries {
            u16 num_hook_entries;
            struct nf_hook_entry        hooks[];
            //const struct nf_hook_ops *orig_ops[]; }
    """

    @classmethod
    def symtab_checks(cls, vmlinux) -> bool:
        hooks_type = ["pointer", "nf_hook_entries"]
        return (
            vmlinux.has_symbol("net_namespace_list")
            and vmlinux.has_type("net_device")
            and vmlinux.get_type("net_device").has_member("nf_hooks_ingress")
            and cls.get_member_type(vmlinux.get_type("net_device"), "nf_hooks_ingress")
            == hooks_type
        )

    def get_hook_ops(self, hook_container, proto_idx, hook_idx):
        nf_hook_entries = hook_container
        if not nf_hook_entries:
            return

        nf_hook_ops_name = self.get_symbol_fullname("nf_hook_ops")
        nf_hook_ops_ptr_arr = self.build_nf_hook_ops_array(nf_hook_entries)
        if not nf_hook_ops_ptr_arr:
            return

        for nf_hook_ops_ptr in nf_hook_ops_ptr_arr:
            nf_hook_ops = nf_hook_ops_ptr.dereference().cast(nf_hook_ops_name)
            yield nf_hook_ops


class Netfilter(interfaces.plugins.PluginInterface):
    """Lists Netfilter hooks."""

    _required_framework_version = (2, 22, 0)

    _version = (2, 0, 0)

    _required_linux_utilities_modules_version = (3, 0, 0)
    _required_linuxutils_gatherers_version = (1, 0, 0)
    _required_linuxnet_version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=cls._required_linuxutils_gatherers_version,
            ),
            requirements.VersionRequirement(
                name="linuxnet",
                component=network.NetSymbols,
                version=cls._required_linuxnet_version,
            ),
        ]

    def _format_fields(self, fields):
        (
            netns,
            proto_name,
            hook_name,
            priority,
            hook_func,
            module_info,
            symbol_name,
            hooked,
        ) = fields

        if module_info:
            module_name = module_info.name
        else:
            module_name = renderers.NotAvailableValue()

        return (
            netns,
            proto_name,
            hook_name,
            priority,
            format_hints.Hex(hook_func),
            module_name,
            symbol_name or renderers.NotAvailableValue(),
            str(hooked),
        )

    def _generator(self):
        kernel_module_name = self.config["kernel"]
        for fields in AbstractNetfilter.run_all(
            context=self.context, kernel_module_name=kernel_module_name
        ):
            yield (0, self._format_fields(fields))

    def run(self):
        headers = [
            ("Net NS", int),
            ("Proto", str),
            ("Hook", str),
            ("Priority", int),
            ("Handler", format_hints.Hex),
            ("Module", str),
            ("Symbol", str),
            ("Is Hooked", str),
        ]
        return renderers.TreeGrid(headers, self._generator())



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Callable, Tuple, List, Dict

from volatility3.framework import interfaces, exceptions, constants, objects, renderers
from volatility3.framework.renderers import format_hints
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.symbols import linux
from volatility3.plugins.linux import lsof
from volatility3.plugins.linux import pslist
from volatility3.framework.symbols.linux import network


vollog = logging.getLogger(__name__)


class SockHandlers(interfaces.configuration.VersionableInterface):
    """Handles several socket families extracting the sockets information."""

    _required_framework_version = (2, 22, 0)
    _version = (4, 0, 0)
    _net_version_required = (1, 0, 0)

    def __init__(self, context, vmlinux_name, task, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vmlinux = context.modules[vmlinux_name]
        self._symbol_table = context.symbol_space[self._vmlinux.symbol_table_name]
        self._task = task

        if not requirements.VersionRequirement.matches_required(
            network.NetSymbols.version, self._net_version_required
        ):
            raise ValueError(
                f"Version mismatch of volatility library NetSymbols version ({network.NetSymbols.version}) and needed version ({self._net_version_required})"
            )

        network.NetSymbols.apply(self._symbol_table)

        try:
            netns_id = task.nsproxy.net_ns.get_inode()
        except AttributeError:
            netns_id = renderers.NotAvailableValue()

        self._netdevices = self._build_network_devices_map(netns_id)

        self._sock_family_handlers = {
            "AF_UNIX": self._unix_sock,
            "AF_INET": self._inet_sock,
            "AF_INET6": self._inet_sock,
            "AF_NETLINK": self._netlink_sock,
            "AF_VSOCK": self._vsock_sock,
            "AF_PACKET": self._packet_sock,
            "AF_XDP": self._xdp_sock,
            "AF_BLUETOOTH": self._bluetooth_sock,
        }

    def _build_network_devices_map(self, netns_id: int) -> Dict:
        """Given a namespace ID it returns a dictionary mapping each network
        interface index (ifindex) to its network interface name:

        Args:
            netns_id: The network namespace ID

        Returns:
            netdevices_map: Mapping network interface index (ifindex) to network
                            interface name
        """
        netdevices_map = {}
        nethead = self._vmlinux.object_from_symbol(symbol_name="net_namespace_list")
        net_symname = self._vmlinux.symbol_table_name + constants.BANG + "net"
        for net in nethead.to_list(net_symname, "list"):
            net_device_symname = (
                self._vmlinux.symbol_table_name + constants.BANG + "net_device"
            )
            for net_dev in net.dev_base_head.to_list(net_device_symname, "dev_list"):
                if (
                    isinstance(netns_id, renderers.NotAvailableValue)
                    or net.get_inode() != netns_id
                ):
                    continue
                dev_name = utility.array_to_string(net_dev.name)
                netdevices_map[net_dev.ifindex] = dev_name
        return netdevices_map

    def process_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str], Dict]:
        """Takes a kernel generic `sock` object and processes it with its respective socket family

        Args:
            sock: Kernel generic `sock` object

        Returns a tuple with:
            sock: The respective kernel's \\*_sock object for that socket family
            sock_stat: A tuple with the source and destination (address and port) along with its state string
            socket_filter: A dictionary with information about the socket filter
        """
        family = sock.get_family()
        socket_filter = {}
        sock_handler = self._sock_family_handlers.get(family)
        if sock_handler:
            try:
                unix_sock, sock_stat = sock_handler(sock)
                self._update_socket_filters_info(sock, socket_filter)

                return unix_sock, sock_stat, socket_filter
            except exceptions.SymbolError as e:
                # Cannot finds the *_sock type in the symbols
                vollog.log(
                    constants.LOGLEVEL_V,
                    "Error processing socket family '%s': %s",
                    family,
                    e,
                )
        else:
            vollog.log(constants.LOGLEVEL_V, "Unsupported family '%s'", family)

        # Even if the sock family is not supported, or the required types
        # are not present in the symbols, we can still show some general
        # information about the socket that may be helpful.
        src_addr = src_port = dst_addr = dst_port = None
        state = sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state

        return sock, sock_stat, socket_filter

    def _update_socket_filters_info(
        self, sock: objects.Pointer, socket_filter: dict
    ) -> None:
        """Get information from the socket and reuseport filters

        Args:
            sock: The kernel sock (sk) struct
            socket_filter: A dictionary with information about the socket filter
        """
        if sock.has_member("sk_filter") and sock.sk_filter:
            sock_filter = sock.sk_filter
            socket_filter["filter_type"] = "socket_filter"
            self._extract_socket_filter_info(sock_filter, socket_filter)

        if sock.has_member("sk_reuseport_cb") and sock.sk_reuseport_cb:
            sock_reuseport_cb = sock.sk_reuseport_cb
            socket_filter["filter_type"] = "reuseport_filter"
            self._extract_socket_filter_info(sock_reuseport_cb, socket_filter)

    def _extract_socket_filter_info(
        self, sock_filter: objects.Pointer, socket_filter: dict
    ) -> None:
        """Get specific information for each type of filter

        Args:
            socket_filter: A dictionary with information about the socket filter
        """
        socket_filter["bpf_filter_type"] = "cBPF"

        if not sock_filter.has_member("prog") or not sock_filter.prog:
            return None

        bpfprog = sock_filter.prog

        bpfprog_type = bpfprog.get_type()
        if not bpfprog_type:
            # kernel < 3.18.140, it's a cBPF filter
            return None

        if bpfprog_type == "BPF_PROG_TYPE_UNSPEC":
            return None  # cBPF filter

        if bpfprog_type != "BPF_PROG_TYPE_SOCKET_FILTER":
            socket_filter["bpf_filter_type"] = f"UNK({bpfprog_type})"
            vollog.warning(f"Unexpected BPF type {bpfprog_type} for a socket")
            return None

        socket_filter["bpf_filter_type"] = "eBPF"
        if not bpfprog.has_member("aux") or not bpfprog.aux:
            return  # kernel < 3.18.140
        bpfprog_aux = bpfprog.aux

        if bpfprog_aux.has_member("id"):
            # `id` member was added to `bpf_prog_aux` in kernels 4.13.16
            socket_filter["bpf_filter_id"] = str(bpfprog_aux.id)
        if bpfprog_aux.has_member("name"):
            # `name` was added to `bpf_prog_aux` in kernels 4.15.18
            bpfprog_name = utility.array_to_string(bpfprog_aux.name)
            if bpfprog_name:
                socket_filter["bpf_filter_name"] = bpfprog_name

    def _unix_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_UNIX socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            unix_sock: The kernel's `unix_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        unix_sock = sock.cast("unix_sock")
        state = unix_sock.get_state()
        src_addr = unix_sock.get_name()
        src_port = unix_sock.get_inode()

        if unix_sock.peer:
            peer = unix_sock.peer.dereference().cast("unix_sock")
            dst_addr = peer.get_name()
            dst_port = peer.get_inode()
        else:
            dst_addr = dst_port = None

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return unix_sock, sock_stat

    def _inet_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_INET/6 socket families

        Args:
            sock: Kernel generic `sock` object

        Returns:
            inet_sock: The kernel's `inet_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        inet_sock = sock.cast("inet_sock")
        src_addr = inet_sock.get_src_addr()
        src_port = inet_sock.get_src_port()
        dst_addr = inet_sock.get_dst_addr()
        dst_port = inet_sock.get_dst_port()
        state = inet_sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return inet_sock, sock_stat

    def _netlink_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_NETLINK socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            netlink_sock: The kernel's `netlink_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        netlink_sock = sock.cast("netlink_sock")

        src_addr = None
        if netlink_sock.groups:
            groups_bitmap = netlink_sock.groups.dereference()
            src_addr = f"groups:0x{groups_bitmap:08x}"

        try:
            # Kernel >= 3.7.10
            src_port = netlink_sock.get_portid()
        except AttributeError:
            src_port = renderers.NotAvailableValue()

        dst_addr = f"group:0x{netlink_sock.dst_group:08x}"
        module = netlink_sock.module
        if module and module.name:
            module_name_str = utility.array_to_string(module.name)
            dst_addr = f"{dst_addr},lkm:{module_name_str}"
        try:
            dst_port = netlink_sock.get_dst_portid()
        except AttributeError:
            dst_port = renderers.NotAvailableValue()

        state = netlink_sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return netlink_sock, sock_stat

    def _vsock_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_VSOCK socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            vsock_sock: The kernel `vsock_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        vsock_sock = sock.cast("vsock_sock")
        src_addr = vsock_sock.local_addr.svm_cid
        src_port = vsock_sock.local_addr.svm_port
        dst_addr = vsock_sock.remote_addr.svm_cid
        dst_port = vsock_sock.remote_addr.svm_port
        state = vsock_sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return vsock_sock, sock_stat

    def _packet_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_PACKET socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            packet_sock: The kernel's `packet_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        packet_sock = sock.cast("packet_sock")
        ifindex = packet_sock.ifindex
        dev_name = self._netdevices.get(ifindex) if ifindex > 0 else "ANY"

        src_addr = dev_name
        src_port = dst_addr = dst_port = None
        state = packet_sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return packet_sock, sock_stat

    def _xdp_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_XDP socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            xdp_sock: The kernel's `xdp_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        xdp_sock = sock.cast("xdp_sock")
        device = xdp_sock.dev
        if not device:
            return None

        src_addr = utility.array_to_string(device.name)
        src_port = dst_addr = dst_port = None

        bpfprog = device.xdp_prog
        if not bpfprog:
            return None

        if not bpfprog.has_member("aux") or not bpfprog.aux:
            return None

        bpfprog_aux = bpfprog.aux
        if bpfprog_aux.has_member("id"):
            # `id` member was added to `bpf_prog_aux` in kernels 4.13
            bpfprog_id = bpfprog_aux.id
            dst_port = f"ebpf_prog_id:{bpfprog_id}"
        if bpfprog_aux.has_member("name"):
            # `name` was added to `bpf_prog_aux` in kernels 4.15
            bpf_name = utility.array_to_string(bpfprog_aux.name)
            if bpf_name:
                dst_addr = f"ebpf_prog_name:{bpf_name}"

        xsk_state = xdp_sock.get_state()
        state = xsk_state.replace("XSK_", "")

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return xdp_sock, sock_stat

    def _bluetooth_sock(
        self, sock: objects.StructType
    ) -> Tuple[objects.StructType, Tuple[str, str, str]]:
        """Handles the AF_BLUETOOTH socket family

        Args:
            sock: Kernel generic `sock` object

        Returns:
            bt_sock: The kernel's `bt_sock` object
            sock_stat: A tuple with the source and destination (address and port) along with its state string
        """
        bt_sock = sock.cast("bt_sock")

        def bt_addr(addr):
            return ":".join(reversed([f"{x:02x}" for x in addr.b]))

        src_addr = src_port = dst_addr = dst_port = None
        bt_protocol = bt_sock.get_protocol()
        if bt_protocol == "HCI":
            if self._vmlinux.has_type("hci_pinfo"):
                pinfo = bt_sock.cast("hci_pinfo")
                if (
                    pinfo.has_member("hdev")
                    and self._vmlinux.has_type("hci_dev")
                    and pinfo.hdev.has_member("dev_name")
                ):
                    src_addr = utility.array_to_string(pinfo.hdev.dev_name)
            else:
                vollog.log(
                    constants.LOGLEVEL_V,
                    "Type definition for 'hci_pinfo' is not available in the symbols",
                )
        elif bt_protocol == "L2CAP":
            if self._vmlinux.has_type("l2cap_pinfo"):
                pinfo = bt_sock.cast("l2cap_pinfo")
                src_addr = bt_addr(pinfo.chan.src)
                dst_addr = bt_addr(pinfo.chan.dst)
                src_port = pinfo.chan.sport
                dst_port = pinfo.chan.psm
            else:
                vollog.log(
                    constants.LOGLEVEL_V,
                    "Type definition for 'l2cap_pinfo' is not available in the symbols",
                )
        elif bt_protocol == "RFCOMM":
            if self._vmlinux.has_type("rfcomm_pinfo"):
                pinfo = bt_sock.cast("rfcomm_pinfo")
                src_addr = bt_addr(pinfo.src)
                dst_addr = bt_addr(pinfo.dst)
                src_port = pinfo.channel
            else:
                vollog.log(
                    constants.LOGLEVEL_V,
                    "Type definition for 'rfcomm_pinfo' is not available in the symbols",
                )
        elif bt_protocol == "SCO":
            if self._vmlinux.has_type("sco_pinfo"):
                pinfo = bt_sock.cast("sco_pinfo")
                src_addr = bt_addr(pinfo.src)
                dst_addr = bt_addr(pinfo.dst)
            else:
                vollog.log(
                    constants.LOGLEVEL_V,
                    "Type definition for 'sco_pinfo' is not available in the symbols",
                )
        else:
            vollog.log(
                constants.LOGLEVEL_V, "Unsupported bluetooth protocol '%s'", bt_protocol
            )

        state = bt_sock.get_state()

        sock_stat = src_addr, src_port, dst_addr, dst_port, state
        return bt_sock, sock_stat


class Sockstat(plugins.PluginInterface):
    """Lists all network connections for all processes."""

    _required_framework_version = (2, 0, 0)
    _version = (3, 0, 4)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="SockHandlers", component=SockHandlers, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="lsof", component=lsof.Lsof, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="linux_net", component=network.NetSymbols, version=(1, 0, 0)
            ),
            requirements.ListRequirement(
                name="pids",
                description="Filter results by process IDs. "
                "It takes the root PID namespace identifiers.",
                element_type=int,
                optional=True,
            ),
            requirements.IntRequirement(
                name="netns",
                description="Filter results by network namespace. "
                "Otherwise, all of them are shown.",
                optional=True,
            ),
        ]

    @classmethod
    def list_sockets(
        cls,
        context: interfaces.context.ContextInterface,
        symbol_table: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ):
        """Returns every single socket descriptor

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of the kernel module on which to operate
            filter_func: A function which takes a task object and returns True if the task should be ignored/filtered

        Yields:
            task: Kernel's task object
            netns_id: Network namespace ID
            fd_num: File descriptor number
            family: Socket family string (AF_UNIX, AF_INET, etc)
            sock_type: Socket type string (STREAM, DGRAM, etc)
            protocol: Protocol string (UDP, TCP, etc)
            sock_fields: A tuple with the \\*_sock object, the sock stats and the extended info dictionary
        """
        vmlinux = context.modules[symbol_table]

        sfop_addr = vmlinux.object_from_symbol("socket_file_ops").vol.offset
        dfop_addr = vmlinux.object_from_symbol("sockfs_dentry_operations").vol.offset

        fd_generator = lsof.Lsof.list_fds(context, vmlinux.name, filter_func)
        for fd_internal in fd_generator:
            fd_num, filp, _full_path = fd_internal.fd_fields
            task = fd_internal.task

            if not (filp.f_op and filp.f_op.is_readable()):
                continue

            if filp.f_op not in (sfop_addr, dfop_addr):
                continue

            dentry = filp.get_dentry()
            if not (dentry and dentry.is_readable()):
                continue

            d_inode = dentry.d_inode
            if not (d_inode and d_inode.is_readable()):
                continue

            socket_alloc = linux.LinuxUtilities.container_of(
                d_inode, "socket_alloc", "vfs_inode", vmlinux
            )
            if not socket_alloc:
                continue
            socket = socket_alloc.socket
            if not (socket.sk and socket.sk.is_readable()):
                continue
            sock = socket.sk.dereference()

            try:
                sock_type = sock.get_type()
                family = sock.get_family()
                sock_handler = SockHandlers(context, vmlinux.name, task)
                sock_fields = sock_handler.process_sock(sock)
            except exceptions.InvalidAddressException:
                continue

            if not sock_fields:
                continue

            child_sock = sock_fields[0]
            protocol = child_sock.get_protocol()

            net = task.nsproxy.net_ns
            try:
                netns_id = net.get_inode()
            except AttributeError:
                netns_id = renderers.NotAvailableValue()

            yield task, netns_id, fd_num, family, sock_type, protocol, sock_fields

    def _format_fields(self, sock_stat, protocol):
        """Prepare the socket fields to be rendered

        Args:
            sock_stat: A tuple with the source and destination (address and port) along with its state string
            protocol: Protocol string (UDP, TCP, etc)

        Returns:
            `sock_stat` and `protocol` formatted.
        """
        sock_stat = [
            renderers.NotAvailableValue() if field is None else str(field)
            for field in sock_stat
        ]
        if protocol is None:
            protocol = renderers.NotAvailableValue()

        return tuple(sock_stat), protocol

    def _generator(self, pids: List[int], netns_id_arg: int, kernel_module_name: str):
        """Enumerate tasks sockets. Each row represents a kernel socket.

        Args:
            pids: List of PIDs to filter. If a empty list or
            netns_id_arg: If a network namespace ID is set, it will only show this namespace.
            symbol_table: The name of the kernel module on which to operate

        Yields:
            netns_id: Network namespace ID
            family: Socket family string (AF_UNIX, AF_INET, etc)
            sock_type: Socket type string (STREAM, DGRAM, etc)
            protocol: Protocol string (UDP, TCP, etc)
            source addr: Source address string
            source port: Source port string (not all of them are int)
            destination addr: Destination address string
            destination port: Destination port (not all of them are int)
            state: State strings (LISTEN, CONNECTED, etc)
            tasks: String with a list of tasks and FDs using a socket. It can also have
                   extended information such as socket filters, bpf info, etc.
        """
        vmlinux = self.context.modules[kernel_module_name]
        symbol_table = self.context.symbol_space[vmlinux.symbol_table_name]
        network.NetSymbols.apply(symbol_table)

        filter_func = pslist.PsList.create_pid_filter(pids)
        socket_generator = self.list_sockets(
            self.context, kernel_module_name, filter_func=filter_func
        )

        for (
            task,
            netns_id,
            fd_num,
            family,
            sock_type,
            protocol,
            sock_fields,
        ) in socket_generator:
            if netns_id_arg and netns_id_arg != netns_id:
                continue

            sock, sock_stat, extended = sock_fields
            sock_stat, protocol = self._format_fields(sock_stat, protocol)

            socket_filter_str = (
                ",".join(f"{k}={v}" for k, v in extended.items())
                if extended
                else renderers.NotAvailableValue()
            )

            task_comm = utility.array_to_string(task.comm)

            fields = (
                netns_id,
                task_comm,
                task.tgid,
                task.pid,
                fd_num,
                format_hints.Hex(sock.vol.offset),
                family,
                sock_type,
                protocol,
                *sock_stat,
                socket_filter_str,
            )

            yield (0, fields)

    def run(self):
        pids = self.config.get("pids")
        netns_id = self.config["netns"]
        kernel_module_name = self.config["kernel"]

        tree_grid_args = [
            ("NetNS", int),
            ("Process Name", str),
            ("PID", int),
            ("TID", int),
            ("FD", int),
            ("Sock Offset", format_hints.Hex),
            ("Family", str),
            ("Type", str),
            ("Proto", str),
            ("Source Addr", str),
            ("Source Port", str),
            ("Destination Addr", str),
            ("Destination Port", str),
            ("State", str),
            ("Filter", str),
        ]

        return renderers.TreeGrid(
            tree_grid_args, self._generator(pids, netns_id, kernel_module_name)
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import random
import string
from typing import Optional, Union

from volatility3.framework import objects, interfaces


class GenericIntelProcess(objects.StructType):
    def _add_process_layer(
        self,
        context: interfaces.context.ContextInterface,
        dtb: Union[int, interfaces.objects.ObjectInterface],
        config_prefix: Optional[str] = None,
        preferred_name: Optional[str] = None,
    ) -> str:
        """Constructs a new layer based on the process's DirectoryTableBase."""

        if config_prefix is None:
            # TODO: Ensure collisions can't happen by verifying the config_prefix is empty
            random_prefix = "".join(
                random.SystemRandom().choice(string.ascii_uppercase + string.digits)
                for _ in range(8)
            )
            config_prefix = interfaces.configuration.path_join(
                "temporary", "_" + random_prefix
            )

        # Figure out a suitable name we can use for the new layer
        if preferred_name is None:
            preferred_name = context.layers.free_layer_name(
                prefix=self.vol.layer_name + "_Process"
            )
        else:
            if preferred_name in context.layers:
                preferred_name = context.layers.free_layer_name(prefix=preferred_name)

        # Copy the parent's config and then make suitable changes
        parent_layer = context.layers[self.vol.layer_name]
        parent_config = parent_layer.build_configuration()
        # It's an intel layer, because we hardwire the "memory_layer" config option
        # FIXME: this could be for other architectures if we don't hardwire this/these values
        parent_config["memory_layer"] = parent_layer.config["memory_layer"]
        parent_config["page_map_offset"] = dtb

        # Set the new configuration and construct the layer
        config_path = interfaces.configuration.path_join(config_prefix, preferred_name)
        context.config.splice(config_path, parent_config)
        new_layer = parent_layer.__class__(
            context, config_path=config_path, name=preferred_name
        )

        # Add the constructed layer and return the name
        context.layers.add_layer(new_layer)
        return preferred_name



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import os
from typing import Dict, List, NamedTuple, Optional, Tuple, Union, cast, Callable

from volatility3.framework import (
    constants,
    exceptions,
    interfaces,
    objects,
    renderers,
    symbols,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import scanners, registry
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import versions
from volatility3.framework.symbols.windows.extensions import services as services_types
from volatility3.plugins.windows import pslist
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


class ServiceBinaryInfo(NamedTuple):
    dll: Union[str, interfaces.renderers.BaseAbsentValue]
    binary: Union[str, interfaces.renderers.BaseAbsentValue]


class SvcScan(interfaces.plugins.PluginInterface):
    """Scans for windows services."""

    _required_framework_version = (2, 0, 0)
    _version = (4, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumeration_method = self.service_scan

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def get_record_tuple(
        cls,
        service_record: interfaces.objects.ObjectInterface,
        binary_info: ServiceBinaryInfo,
    ):
        return (
            format_hints.Hex(service_record.vol.offset),
            service_record.Order,
            service_record.get_pid(),
            service_record.Start.description,
            service_record.State.description,
            service_record.get_type(),
            service_record.get_name(),
            service_record.get_display(),
            service_record.get_binary(),
            binary_info.binary,
            binary_info.dll,
        )

    # These checks must be completed from newest -> oldest OS version.
    _win_version_file_map: List[Tuple[versions.OsDistinguisher, bool, str]] = [
        (versions.is_win10_25398_or_later, True, "services-win10-25398-x64"),
        (versions.is_win10_19041_or_later, True, "services-win10-19041-x64"),
        (versions.is_win10_19041_or_later, False, "services-win10-19041-x86"),
        (versions.is_win10_18362_or_later, True, "services-win10-18362-x64"),
        (versions.is_win10_18362_or_later, False, "services-win10-18362-x86"),
        (versions.is_win10_17763_or_later, False, "services-win10-17763-x86"),
        (versions.is_win10_16299_or_later, True, "services-win10-16299-x64"),
        (versions.is_win10_16299_or_later, False, "services-win10-16299-x86"),
        (versions.is_win10_15063, True, "services-win10-15063-x64"),
        (versions.is_win10_15063, False, "services-win10-15063-x86"),
        (versions.is_win10_up_to_15063, True, "services-win8-x64"),
        (versions.is_win10_up_to_15063, False, "services-win8-x86"),
        (versions.is_windows_8_or_later, True, "services-win8-x64"),
        (versions.is_windows_8_or_later, True, "services-win8-x86"),
        (versions.is_vista_or_later, True, "services-vista-x64"),
        (versions.is_vista_or_later, False, "services-vista-x86"),
        (versions.is_windows_xp, False, "services-xp-x86"),
        (versions.is_xp_or_2003, True, "services-xp-2003-x64"),
    ]

    @staticmethod
    def _create_service_table(
        context: interfaces.context.ContextInterface,
        symbol_table_name: str,
        config_path: str,
    ) -> str:
        """Constructs a symbol table containing the symbols for services
        depending upon the operating system in use.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of the table containing the kernel symbols
            config_path: The configuration path for any settings required by the new table

        Returns:
            A symbol table containing the symbols necessary for services
        """
        native_types = context.symbol_space[symbol_table_name].natives
        is_64bit = symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=symbol_table_name
        )

        try:
            symbol_filename = next(
                filename
                for version_check, for_64bit, filename in SvcScan._win_version_file_map
                if is_64bit == for_64bit
                and version_check(context=context, symbol_table=symbol_table_name)
            )
        except StopIteration:
            raise NotImplementedError("This version of Windows is not supported!")

        return intermed.IntermediateSymbolTable.create(
            context,
            config_path,
            os.path.join("windows", "services"),
            symbol_filename,
            class_types=services_types.class_types,
            native_types=native_types,
        )

    @staticmethod
    def _get_service_key(
        context, config_path: str, kernel_module_name: str
    ) -> Optional[objects.StructType]:
        for hive in hivelist.HiveList.list_hives(
            context=context,
            base_config_path=interfaces.configuration.path_join(
                config_path, "hivelist"
            ),
            kernel_module_name=kernel_module_name,
            filter_string="machine\\system",
        ):
            # Get ControlSet\Services.
            try:
                return cast(
                    objects.StructType, hive.get_key(r"CurrentControlSet\Services")
                )
            except (
                KeyError,
                exceptions.InvalidAddressException,
                registry.RegistryException,
            ):
                try:
                    return cast(
                        objects.StructType, hive.get_key(r"ControlSet001\Services")
                    )
                except (
                    KeyError,
                    exceptions.InvalidAddressException,
                    registry.RegistryException,
                ):
                    vollog.log(
                        constants.LOGLEVEL_VVVV,
                        "Could not retrieve any control set from SYSTEM hive",
                    )

        return None

    @staticmethod
    def _get_service_dll(
        service_key,
    ) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        try:
            param_key = next(
                key
                for key in service_key.get_subkeys()
                if key.get_name() == "Parameters"
            )
            return (
                next(
                    val
                    for val in param_key.get_values()
                    if val.get_name() == "ServiceDll"
                )
                .decode_data()
                .decode("utf-16")
                .rstrip("\x00")
            )

        except UnicodeDecodeError:
            return renderers.UnparsableValue()
        except StopIteration:
            return renderers.UnreadableValue()

    @staticmethod
    def _get_service_binary(
        service_key,
    ) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        try:
            return (
                next(
                    val
                    for val in service_key.get_values()
                    if val.get_name() == "ImagePath"
                )
                .decode_data()
                .decode("utf-16")
                .rstrip("\x00")
            )

        except UnicodeDecodeError:
            return renderers.UnparsableValue()
        except StopIteration:
            return renderers.UnreadableValue()

    @staticmethod
    def _get_service_binary_map(
        services_key: interfaces.objects.ObjectInterface,
    ) -> Dict[str, ServiceBinaryInfo]:
        services = services_key.get_subkeys()
        return {
            service_key.get_name(): ServiceBinaryInfo(
                SvcScan._get_service_dll(service_key),
                SvcScan._get_service_binary(service_key),
            )
            for service_key in services
        }

    @classmethod
    def enumerate_vista_or_later_header(
        cls,
        context,
        service_table_name,
        service_binary_dll_map,
        proc_layer_name,
        offset,
    ):
        if offset % 8:
            return

        service_header = context.object(
            service_table_name + constants.BANG + "_SERVICE_HEADER",
            offset=offset,
            layer_name=proc_layer_name,
        )

        if not service_header.is_valid():
            return

        # since we walk the s-list backwards, if we've seen
        # an object, then we've also seen all objects that
        # exist before it, thus we can break at that time.
        for service_record in service_header.ServiceRecord.traverse():
            service_info = service_binary_dll_map.get(
                service_record.get_name(),
                ServiceBinaryInfo(
                    renderers.UnreadableValue(), renderers.UnreadableValue()
                ),
            )
            yield cls.get_record_tuple(service_record, service_info)

    @classmethod
    def service_scan(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        service_table_name: str,
        service_binary_dll_map,
        filter_func,
    ):
        kernel = context.modules[kernel_module_name]

        relative_tag_offset = context.symbol_space.get_type(
            service_table_name + constants.BANG + "_SERVICE_RECORD"
        ).relative_child_offset("Tag")

        is_vista_or_later = versions.is_vista_or_later(
            context=context, symbol_table=kernel.symbol_table_name
        )

        if is_vista_or_later:
            service_tag = b"serH"
        else:
            service_tag = b"sErv"

        seen = []

        for task in pslist.PsList.list_processes(
            context,
            kernel_module_name=kernel_module_name,
            filter_func=filter_func,
        ):
            proc_id = "Unknown"
            try:
                proc_id = task.UniqueProcessId
                proc_layer_name = task.add_process_layer()
            except exceptions.InvalidAddressException as excp:
                vollog.debug(
                    f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
                )
                continue

            process_layer = context.layers[proc_layer_name]

            # get process sections for scanning
            sections = []
            for vad in task.get_vad_root().traverse():
                base = vad.get_start()
                if vad.get_size():
                    sections.append((base, vad.get_size()))

            for offset in process_layer.scan(
                context=context,
                scanner=scanners.BytesScanner(needle=service_tag),
                sections=sections,
            ):
                if not is_vista_or_later:
                    service_record = context.object(
                        service_table_name + constants.BANG + "_SERVICE_RECORD",
                        offset=offset - relative_tag_offset,
                        layer_name=proc_layer_name,
                    )

                    if not service_record.is_valid():
                        continue

                    service_info = service_binary_dll_map.get(
                        service_record.get_name(),
                        ServiceBinaryInfo(
                            renderers.UnreadableValue(), renderers.UnreadableValue()
                        ),
                    )
                    yield cls.get_record_tuple(service_record, service_info)
                else:
                    for service_record in cls.enumerate_vista_or_later_header(
                        context,
                        service_table_name,
                        service_binary_dll_map,
                        proc_layer_name,
                        offset,
                    ):
                        if service_record in seen:
                            break
                        seen.append(service_record)
                        yield service_record

    @classmethod
    def get_prereq_info(
        cls, context, config_path: str, kernel_module_name: str
    ) -> Tuple[str, Dict, Callable]:
        """
        Data structures and information needed to analyze service information
        """

        kernel = context.modules[kernel_module_name]

        service_table_name = cls._create_service_table(
            context, kernel.symbol_table_name, config_path
        )

        services_key = cls._get_service_key(context, config_path, kernel_module_name)

        service_binary_dll_map = (
            cls._get_service_binary_map(services_key)
            if services_key is not None
            else {}
        )

        filter_func = pslist.PsList.create_name_filter(["services.exe"])

        return service_table_name, service_binary_dll_map, filter_func

    def _generator(self):
        service_table_name, service_binary_dll_map, filter_func = self.get_prereq_info(
            self.context, self.config_path, self.config["kernel"]
        )

        for record in self._enumeration_method(
            self.context,
            self.config["kernel"],
            service_table_name,
            service_binary_dll_map,
            filter_func,
        ):
            yield (0, record)

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Order", int),
                ("PID", int),
                ("Start", str),
                ("State", str),
                ("Type", str),
                ("Name", str),
                ("Display", str),
                ("Binary", str),
                ("Binary (Registry)", str),
                ("Dll", str),
            ],
            self._generator(),
        )



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import datetime
import dataclasses
import contextlib
from typing import Any, Callable, Iterable, List, Optional

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.linux.extensions import elf
from volatility3.plugins import timeliner
from volatility3.plugins.linux import elfs


@dataclasses.dataclass
class TaskFields:
    offset: int
    user_pid: int
    user_tid: int
    user_ppid: int
    name: str
    uid: Optional[int]
    gid: Optional[int]
    euid: Optional[int]
    egid: Optional[int]
    creation_time: Optional[datetime.datetime]


class PsList(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists the processes present in a particular linux memory image."""

    _required_framework_version = (2, 13, 0)
    _version = (4, 1, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="elfs", component=elfs.Elfs, version=(2, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.BooleanRequirement(
                name="threads",
                description="Include user threads",
                optional=True,
                default=False,
            ),
            requirements.BooleanRequirement(
                name="decorate_comm",
                description="Show `user threads` comm in curly brackets, and `kernel threads` comm in square brackets",
                optional=True,
                default=False,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed processes",
                optional=True,
                default=False,
            ),
        ]

    @classmethod
    def create_pid_filter(
        cls, pid_list: Optional[List[int]] = None
    ) -> Callable[[Any], bool]:
        """Constructs a filter function for process IDs.

        Args:
            pid_list: List of process IDs that are acceptable (or None if all are acceptable)

        Returns:
            Function which, when provided a process object, returns True if the process is to be filtered out of the list
        """
        pid_list = pid_list or []
        filter_list = [x for x in pid_list if x is not None]
        if filter_list:

            def filter_func(x):
                return x.pid not in filter_list

            return filter_func
        else:
            return lambda _: False

    @classmethod
    def get_task_fields(
        cls, task: interfaces.objects.ObjectInterface, decorate_comm: bool = False
    ) -> TaskFields:
        """Extract the fields needed for the final output

        Args:
            task: A task object from where to get the fields.
            decorate_comm: If True, it decorates the comm string of user threads in curly brackets,
                           and of Kernel threads in square brackets.
                           Defaults to False.
        Returns:
            A TaskFields object with the fields to show in the plugin output.
        """
        name = utility.array_to_string(task.comm)
        if decorate_comm:
            if task.is_kernel_thread:
                name = f"[{name}]"
            elif task.is_user_thread:
                name = f"{{{name}}}"

        # This function may be called with a partially initialized/uninitialized task.
        # Ensure it always returns a valid TaskFields object, ready for use in a plugin.
        valid_cred = task.cred and task.cred.is_readable()
        creation_time = None
        with contextlib.suppress(Exception):
            creation_time = task.get_create_time()

        return TaskFields(
            offset=task.vol.offset,
            user_pid=task.tgid,
            user_tid=task.pid,
            user_ppid=task.get_parent_pid(),
            name=name,
            uid=task.cred.uid if valid_cred else None,
            gid=task.cred.gid if valid_cred else None,
            euid=task.cred.euid if valid_cred else None,
            egid=task.cred.egid if valid_cred else None,
            creation_time=creation_time,
        )

    def _get_file_output(self, task: interfaces.objects.ObjectInterface) -> str:
        """Extract the elf for the process if requested
        Args:
            task: A task object to extract from.
        Returns:
            A string showing the results of the extraction, either
            the filename used or an error.
        """
        elf_table_name = intermed.IntermediateSymbolTable.create(
            self.context,
            self.config_path,
            "linux",
            "elf",
            class_types=elf.class_types,
        )
        proc_layer_name = task.add_process_layer()
        if not proc_layer_name:
            # if we can't build a proc layer we can't
            # extract the elf
            return renderers.NotApplicableValue()
        else:
            # Find the vma that belongs to the main ELF of the process
            file_output = "Error outputting file"
            for v in task.mm.get_vma_iter():
                if v.vm_start == task.mm.start_code:
                    file_handle = elfs.Elfs.elf_dump(
                        self.context,
                        proc_layer_name,
                        elf_table_name,
                        v,
                        task,
                        self.open,
                    )
                    if file_handle:
                        file_output = str(file_handle.preferred_filename)
                        file_handle.close()
                    break
            else:
                file_output = "VMA start matching task start_code not found"
        return file_output

    @staticmethod
    def _format_cred(cred):
        return renderers.NotAvailableValue() if cred is None else cred

    def _generator(
        self,
        pid_filter: Callable[[Any], bool],
        include_threads: bool = False,
        decorate_comm: bool = False,
        dump: bool = False,
    ):
        """Generates the tasks list.

        Args:
            pid_filter: A function which takes a process object and returns True if the process should be ignored/filtered
            include_threads: If True, the output will also show the user threads
                             If False, only the thread group leaders will be shown
                             Defaults to False.
            decorate_comm: If True, it decorates the comm string of
                            - User threads: in curly brackets,
                            - Kernel threads: in square brackets
                           Defaults to False.
            dump: If True, the main executable of the process is written to a file
                  Defaults to False.
        Yields:
            Each rows
        """
        for task in self.list_tasks(
            self.context, self.config["kernel"], pid_filter, include_threads
        ):
            if dump:
                file_output = self._get_file_output(task)
            else:
                file_output = "Disabled"

            task_fields = self.get_task_fields(task, decorate_comm)

            task_uid = self._format_cred(task_fields.uid)
            task_gid = self._format_cred(task_fields.gid)
            task_euid = self._format_cred(task_fields.euid)
            task_egid = self._format_cred(task_fields.egid)

            yield (
                0,
                (
                    format_hints.Hex(task_fields.offset),
                    task_fields.user_pid,
                    task_fields.user_tid,
                    task_fields.user_ppid,
                    task_fields.name,
                    task_uid,
                    task_gid,
                    task_euid,
                    task_egid,
                    task_fields.creation_time or renderers.NotAvailableValue(),
                    file_output,
                ),
            )

    @classmethod
    def list_tasks(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
        include_threads: bool = False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the tasks in the primary layer.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a process object and returns True if the process should be ignored/filtered
            include_threads: If True, it will also return user threads.
        Yields:
            Task objects
        """
        vmlinux = context.modules[vmlinux_module_name]

        init_task = vmlinux.object_from_symbol(symbol_name="init_task")

        # Note that the init_task itself is not yielded, since "ps" also never shows it.
        seen = set()
        for forward in (True, False):
            for task in init_task.tasks.to_list(
                symbol_type=init_task.vol.type_name,
                member="tasks",
                forward=forward,
            ):
                if task.vol.offset in seen:
                    continue
                seen.add(task.vol.offset)

                if not task.is_valid():
                    continue

                if filter_func(task):
                    continue

                yield task

                if include_threads:
                    yield from task.get_threads()

    def run(self):
        pids = self.config.get("pid")
        include_threads = self.config.get("threads")
        decorate_comm = self.config.get("decorate_comm")
        dump = self.config.get("dump")
        filter_func = self.create_pid_filter(pids)

        columns = [
            ("OFFSET (V)", format_hints.Hex),
            ("PID", int),
            ("TID", int),
            ("PPID", int),
            ("COMM", str),
            ("UID", int),
            ("GID", int),
            ("EUID", int),
            ("EGID", int),
            ("CREATION TIME", datetime.datetime),
            ("File output", str),
        ]
        return renderers.TreeGrid(
            columns, self._generator(filter_func, include_threads, decorate_comm, dump)
        )

    def generate_timeline(self):
        pids = self.config.get("pid")
        filter_func = self.create_pid_filter(pids)
        for task in self.list_tasks(
            self.context, self.config["kernel"], filter_func, include_threads=True
        ):
            task_fields = self.get_task_fields(task)
            description = f"Process {task_fields.user_pid}/{task_fields.user_tid} {task_fields.name} ({task_fields.offset})"

            yield (
                description,
                timeliner.TimeLinerType.CREATED,
                task_fields.creation_time,
            )



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

"""Functions that read AVML files.

The user of the file doesn't have to worry about the compression,
but random access is not allowed."""

import ctypes
import logging
import struct
from typing import Tuple, List, Optional

from volatility3.framework import exceptions, interfaces, constants
from volatility3.framework.layers import segmented

vollog = logging.getLogger(__name__)

try:
    # TODO: Find library for windows if needed
    try:
        # Linux
        lib_snappy = ctypes.cdll.LoadLibrary("libsnappy.so.1")
    except OSError:
        lib_snappy = None

    try:
        if not lib_snappy:
            # macOS
            lib_snappy = ctypes.cdll.LoadLibrary("libsnappy.1.dylib")
    except OSError:
        lib_snappy = None

    try:
        if not lib_snappy:
            # Windows 64
            lib_snappy = ctypes.cdll.LoadLibrary("snappy64")
    except OSError:
        lib_snappy = None

    if not lib_snappy:
        # Windows 32
        lib_snappy = ctypes.cdll.LoadLibrary("snappy32")

    __snappy_uncompress = lib_snappy.snappy_uncompress
    __snappy_uncompressed_length = lib_snappy.snappy_uncompressed_length

    HAS_SNAPPY = True
except (AttributeError, OSError):
    HAS_SNAPPY = False


class SnappyException(exceptions.VolatilityException):
    pass


def uncompress(s):
    """Uncompress a snappy compressed string."""
    ulen = ctypes.c_int(0)
    cresult = __snappy_uncompressed_length(s, len(s), ctypes.byref(ulen))
    if cresult != 0:
        raise SnappyException(f"Error in snappy_uncompressed_length: {cresult}")
    ubuf = ctypes.create_string_buffer(ulen.value)
    cresult = __snappy_uncompress(s, len(s), ubuf, ctypes.byref(ulen))
    if cresult != 0:
        raise SnappyException(f"Error in snappy_uncompress: {cresult}")
    return ubuf.raw


class AVMLLayer(segmented.NonLinearlySegmentedLayer):
    """A Lime format TranslationLayer.

    Lime is generally used to store physical memory images where there
    are large holes in the physical layer
    """

    def __init__(self, *args, **kwargs):
        self._compressed = {}
        super().__init__(*args, **kwargs)

    @classmethod
    def _check_header(cls, layer: interfaces.layers.DataLayerInterface):
        header_structure = "<II"
        magic, version = struct.unpack(
            header_structure,
            layer.read(layer.minimum_address, struct.calcsize(header_structure)),
        )
        if magic not in [0x4C4D5641] or version != 2:
            raise exceptions.LayerException("File not in AVML format")
        if not HAS_SNAPPY:
            vollog.warning(
                "AVML file detected, but snappy library could not be found\n"
                "Please install the snappy from your distribution or https://google.github.io/snappy/."
            )
            raise exceptions.LayerException(
                "AVML format dependencies not satisfied (snappy)"
            )

    def _load_segments(self) -> None:
        base_layer = self.context.layers[self._base_layer]
        offset = base_layer.minimum_address
        while offset + 4 < base_layer.maximum_address:
            avml_header_structure = "<IIQQQ"
            avml_header_size = struct.calcsize(avml_header_structure)
            avml_header_data = base_layer.read(offset, avml_header_size)
            magic, version, start, end, padding = struct.unpack(
                avml_header_structure, avml_header_data
            )

            if magic not in [0x4C4D5641] or version != 2:
                raise exceptions.LayerException("File not completely in AVML format")
            chunk_data = base_layer.read(
                offset + avml_header_size,
                min(
                    end - start,
                    base_layer.maximum_address - (offset + avml_header_size),
                ),
            )
            segments, consumed = self._read_snappy_frames(chunk_data, end - start)
            # The returned segments are accurate the chunk_data that was passed in, but needs shifting
            for thing, mapped_offset, size, mapped_size, compressed in segments:
                self._segments.append(
                    (
                        thing + start,
                        offset + mapped_offset + avml_header_size,
                        size,
                        mapped_size,
                    )
                )
                self._compressed[offset + mapped_offset + avml_header_size] = compressed

            # TODO: Check whatever the remaining 8 bytes are
            offset += avml_header_size + consumed + 8

    def _read_snappy_frames(
        self, data: bytes, expected_length: int
    ) -> Tuple[List[Tuple[int, int, int, int, bool]], int]:
        """
        Reads a framed-format snappy stream

        Args:
            data: The stream to read
            expected_length: How big the decompressed stream is expected to be (termination limit)

        Returns:
            (offset, mapped_offset, length, mapped_length, compressed) relative to the data chunk (ie, not relative to the file start)
        """
        segments = []
        decompressed_len = 0
        offset = 0
        crc_len = 4
        frame_header_struct = "<L"
        frame_header_len = struct.calcsize(frame_header_struct)
        while decompressed_len <= expected_length:
            if offset + frame_header_len < len(data):
                frame_header = data[offset : offset + frame_header_len]
                frame_header_val = struct.unpack("<L", frame_header)[0]
                frame_type, frame_size = frame_header_val & 0xFF, frame_header_val >> 8
                if frame_type == 0xFF:
                    if (
                        data[
                            offset + frame_header_len : offset
                            + frame_header_len
                            + frame_size
                        ]
                        != b"sNaPpY"
                    ):
                        raise ValueError(f"Snappy header missing at offset: {offset}")
                elif frame_type in [0x00, 0x01]:
                    # CRC + (Un)compressed data
                    mapped_start = offset + frame_header_len
                    # frame_crc = data[mapped_start: mapped_start + crc_len]
                    frame_data = data[
                        mapped_start + crc_len : mapped_start + frame_size
                    ]
                    if frame_type == 0x00:
                        # Compressed data
                        frame_data = uncompress(frame_data)
                    # TODO: Verify CRC
                    segments.append(
                        (
                            decompressed_len,
                            mapped_start + crc_len,
                            len(frame_data),
                            frame_size - crc_len,
                            frame_type == 0x00,
                        )
                    )
                    decompressed_len += len(frame_data)
                elif frame_type in range(0x2, 0x80):
                    # Unskippable
                    raise exceptions.LayerException(
                        f"Unskippable chunk of type {frame_type} found: {offset}"
                    )
                offset += frame_header_len + frame_size
        return segments, offset

    def _decode_data(
        self, data: bytes, mapped_offset: int, offset: int, output_length: int
    ) -> bytes:
        start_offset, _, _, _ = self._find_segment(offset)
        if self._compressed[mapped_offset]:
            decoded_data = uncompress(data)
        else:
            decoded_data = data
        decoded_data = decoded_data[offset - start_offset :]
        decoded_data = decoded_data[:output_length]
        return decoded_data


class AVMLStacker(interfaces.automagic.StackerLayerInterface):
    stack_order = 10

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        try:
            AVMLLayer._check_header(context.layers[layer_name])
        except exceptions.LayerException:
            return None
        new_name = context.layers.free_layer_name("AVMLLayer")
        context.config[interfaces.configuration.path_join(new_name, "base_layer")] = (
            layer_name
        )
        return AVMLLayer(context, new_name, new_name)



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Dict, Iterator, List, Optional

from volatility3.framework import constants, exceptions, interfaces, renderers, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist, psscan

vollog = logging.getLogger(__name__)


class Handles(interfaces.plugins.PluginInterface):
    """Lists process open handles."""

    _required_framework_version = (2, 0, 0)
    _version = (4, 0, 0)

    LEVEL_MASK = 7

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="psscan", component=psscan.PsScan, version=(2, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.IntRequirement(
                name="offset",
                description="Process offset in the physical address space",
                optional=True,
            ),
        ]

    @classmethod
    def _get_item(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        handle_table_entry: interfaces.objects.ObjectInterface,
        handle_value: int,
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """
        Given  a handle table entry (_HANDLE_TABLE_ENTRY) structure from a
        process' handle table, determine where the corresponding object's
        _OBJECT_HEADER can be found, and construct and return the _OBJECT_HEADER
        """

        kernel = context.modules[kernel_module_name]

        virtual = kernel.layer_name

        try:
            # before windows 7
            if not context.layers[virtual].is_valid(handle_table_entry.Object):
                return None
            fast_ref = handle_table_entry.Object.cast("_EX_FAST_REF")

            try:
                object_header = fast_ref.dereference().cast("_OBJECT_HEADER")
            except exceptions.InvalidAddressException:
                return None

            object_header.GrantedAccess = handle_table_entry.GrantedAccess
        except AttributeError:
            # starting with windows 8
            is_64bit = symbols.symbol_table_is_64bit(
                context=context, symbol_table_name=kernel.symbol_table_name
            )

            if is_64bit:
                try:
                    pointer_bits = handle_table_entry.ObjectPointerBits
                except exceptions.InvalidAddressException:
                    return None

                if pointer_bits == 0:
                    return None

                offset = pointer_bits << 4

            else:
                try:
                    info_table = handle_table_entry.InfoTable
                except exceptions.InvalidAddressException:
                    return None

                if info_table == 0:
                    return None

                offset = info_table & ~7

            # print("LowValue: {0:#x} Magic: {1:#x} Offset: {2:#x}".format(handle_table_entry.InfoTable, magic, offset))
            object_header = context.object(
                kernel.symbol_table_name + constants.BANG + "_OBJECT_HEADER",
                virtual,
                offset=offset,
            )
            try:
                object_header.GrantedAccess = handle_table_entry.GrantedAccessBits
            except exceptions.InvalidAddressException:
                return None

        object_header.HandleValue = handle_value
        return object_header

    @classmethod
    def get_type_map(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Dict[int, str]:
        """List the executive object types (_OBJECT_TYPE) using the
        ObTypeIndexTable or ObpObjectTypes symbol (differs per OS). This method
        will be necessary for determining what type of object we have given an
        object header.

        Note:
            The object type index map was hard coded into profiles in previous versions of volatility.
            It is now generated dynamically.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols

        Returns:
            A mapping of type indices to type names
        """

        type_map: Dict[int, str] = {}

        ntkrnlmp = context.modules[kernel_module_name]

        try:
            table_addr = ntkrnlmp.get_symbol("ObTypeIndexTable").address
        except exceptions.SymbolError:
            table_addr = ntkrnlmp.get_symbol("ObpObjectTypes").address

        trans_layer = context.layers[ntkrnlmp.layer_name]

        if not trans_layer.is_valid(ntkrnlmp.offset + table_addr):
            return type_map

        ptrs = ntkrnlmp.object(
            object_type="array",
            offset=table_addr,
            subtype=ntkrnlmp.get_type("pointer"),
            count=100,
        )

        for i, ptr in enumerate(ptrs):  # type: ignore
            # the first entry in the table is always null. break the
            # loop when we encounter the first null entry after that
            if i > 0 and ptr == 0:
                break

            try:
                objt = ptr.dereference().cast(
                    ntkrnlmp.symbol_table_name + constants.BANG + "_OBJECT_TYPE"
                )
                type_name = objt.Name.String
            except exceptions.InvalidAddressException:
                vollog.log(
                    constants.LOGLEVEL_VVV,
                    f"Cannot access _OBJECT_HEADER Name at {ptr.vol.offset:#x}",
                )
                continue

            type_map[i] = type_name

        return type_map

    @classmethod
    def find_cookie(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Find the ObHeaderCookie value (if it exists)"""

        kernel = context.modules[kernel_module_name]

        try:
            symbol_offset = kernel.get_symbol("ObHeaderCookie").address
        except exceptions.SymbolError:
            vollog.debug('Unable to get symbol information for "ObHeaderCookie"')
            return None

        return kernel.object(
            "unsigned int",
            offset=symbol_offset,
        )

    @classmethod
    def _make_handle_array(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        offset: int,
        level: int,
        depth: int = 0,
    ) -> Iterator[interfaces.objects.ObjectInterface]:
        """
        Parses a process' handle table by constructing an array of
        `_HANDLE_TABLE_ENTRY` structures at the given offset, and yields valid
        handle table entries, going as deep into the table "levels" as
        necessary.
        """

        kernel = context.modules[kernel_module_name]

        if level > 0:
            subtype = kernel.get_type("pointer")
            count = 0x1000 / subtype.size
        else:
            subtype = kernel.get_type("_HANDLE_TABLE_ENTRY")
            count = 0x1000 / subtype.size

        if not context.layers[kernel.layer_name].is_valid(offset):
            return None

        table = kernel.object(
            object_type="array",
            offset=offset,
            subtype=subtype,
            count=int(count),
            absolute=True,
        )

        layer_object = context.layers[kernel.layer_name]
        masked_offset = offset & layer_object.maximum_address

        for i in range(len(table)):
            try:
                entry = table[i]
            except exceptions.InvalidAddressException:
                vollog.debug(f"Failed to get handle table entry at index {i}")
                continue
            # This triggered a backtrace in many testing samples
            # in the level == 0 path
            # The code above this calls `is_valid` on the `offset`
            # It is sent but then does not validate `entry` before
            # sending it to `_get_item`
            if not context.layers[kernel.layer_name].is_valid(entry.vol.offset):
                continue

            if level > 0:
                yield from cls._make_handle_array(
                    context, kernel_module_name, entry, level - 1, depth
                )
                depth += 1
            else:
                handle_multiplier = 4
                handle_level_base = depth * count * handle_multiplier

                handle_value = (
                    (entry.vol.offset - masked_offset)
                    / (subtype.size / handle_multiplier)
                ) + handle_level_base

                item = cls._get_item(context, kernel_module_name, entry, handle_value)

                if item is None:
                    continue

                try:
                    if item.TypeIndex != 0x0:
                        yield item
                except AttributeError:
                    if item.Type.Name:
                        yield item
                except exceptions.InvalidAddressException:
                    continue

    @classmethod
    def handles(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        handle_table: interfaces.objects.ObjectInterface,
    ) -> Iterator[interfaces.objects.ObjectInterface]:
        """
        Takes a context, kernel module name, and handle table structure
        (_HANDLE_TABLE), and yields _HANDLE_TABLE_ENTRY structures from the
        handle table.
        """
        try:
            TableCode = handle_table.TableCode & ~cls.LEVEL_MASK
            table_levels = handle_table.TableCode & cls.LEVEL_MASK
        except exceptions.InvalidAddressException:
            vollog.log(
                constants.LOGLEVEL_VVV,
                "Handle table parsing was aborted due to an invalid address exception",
            )
            return None

        yield from cls._make_handle_array(
            context, kernel_module_name, TableCode, table_levels
        )

    def _generator(self, procs):
        type_map = self.get_type_map(
            context=self.context, kernel_module_name=self.config["kernel"]
        )

        cookie = self.find_cookie(
            context=self.context, kernel_module_name=self.config["kernel"]
        )

        for proc in procs:
            try:
                object_table = proc.ObjectTable
            except exceptions.InvalidAddressException:
                vollog.log(
                    constants.LOGLEVEL_VVV,
                    f"Cannot access _EPROCESS.ObjectType at {proc.vol.offset:#x}",
                )
                continue

            process_name = utility.array_to_string(proc.ImageFileName)

            for entry in self.handles(
                self.context, self.config["kernel"], object_table
            ):
                try:
                    obj_type = entry.get_object_type(type_map, cookie)
                    if obj_type is None:
                        continue
                    if obj_type == "File":
                        item = entry.Body.cast("_FILE_OBJECT")
                        obj_name = item.file_name_with_device()
                    elif obj_type == "Process":
                        item = entry.Body.cast("_EPROCESS")
                        obj_name = f"{utility.array_to_string(item.ImageFileName)} Pid {item.UniqueProcessId}"
                    elif obj_type == "Thread":
                        item = entry.Body.cast("_ETHREAD")
                        obj_name = (
                            f"Tid {item.Cid.UniqueThread} Pid {item.Cid.UniqueProcess}"
                        )
                    elif obj_type == "Key":
                        item = entry.Body.cast("_CM_KEY_BODY")
                        obj_name = item.get_full_key_name()
                    else:
                        try:
                            obj_name = entry.NameInfo.Name.String
                        except (ValueError, exceptions.InvalidAddressException):
                            obj_name = None

                except exceptions.InvalidAddressException:
                    vollog.log(
                        constants.LOGLEVEL_VVV,
                        f"Cannot access _OBJECT_HEADER at {entry.vol.offset:#x}",
                    )
                    continue

                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        process_name,
                        format_hints.Hex(entry.Body.vol.offset),
                        format_hints.Hex(entry.HandleValue),
                        obj_type,
                        format_hints.Hex(entry.GrantedAccess),
                        obj_name or renderers.NotAvailableValue(),
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        kernel = self.context.modules[self.config["kernel"]]

        if self.config["offset"]:
            procs = psscan.PsScan.scan_processes(
                self.context,
                self.config["kernel"],
                filter_func=psscan.PsScan.create_offset_filter(
                    self.context,
                    kernel.layer_name,
                    self.config["offset"],
                ),
            )
        else:
            procs = pslist.PsList.list_processes(
                context=self.context,
                kernel_module_name=self.config["kernel"],
                filter_func=filter_func,
            )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Offset", format_hints.Hex),
                ("HandleValue", format_hints.Hex),
                ("Type", str),
                ("GrantedAccess", format_hints.Hex),
                ("Name", str),
            ],
            self._generator(procs=procs),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Volatility 3 Constants.

Stores all the constant values that are generally fixed throughout
volatility This includes default scanning block sizes, etc.
"""

import enum
import os.path
import sys
import warnings
from typing import Callable, Optional

from volatility3.framework.constants import linux as linux
from volatility3.framework.constants import windows as windows
from volatility3.framework.constants._version import (
    PACKAGE_VERSION as PACKAGE_VERSION,
    VERSION_MAJOR as VERSION_MAJOR,
    VERSION_MINOR as VERSION_MINOR,
    VERSION_PATCH as VERSION_PATCH,
    VERSION_SUFFIX as VERSION_SUFFIX,
)

REQUIRED_PYTHON_VERSION = (3, 8, 0)

CACHE_PATH = os.path.join(
    os.environ.get("XDG_CACHE_HOME") or os.path.join(os.path.expanduser("~"), ".cache"),
    "volatility3",
)
"""Default path to store cached data"""

if sys.platform == "win32":
    CACHE_PATH = os.path.realpath(
        os.path.join(os.environ.get("APPDATA", os.path.expanduser("~")), "volatility3")
    )
os.makedirs(CACHE_PATH, exist_ok=True)

PLUGINS_PATH = [
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "plugins")),
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "plugins")),
]
"""Default list of paths to load plugins from (volatility3/plugins and volatility3/framework/plugins)"""

SYMBOL_BASEPATHS = [
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "symbols")),
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "symbols")),
    os.path.abspath(
        os.path.join(CACHE_PATH, "symbols")
    ),  # User cache fallback for automatically downloaded temporary symbols
]
"""Default list of paths to load symbols from (volatility3/symbols and volatility3/framework/symbols)"""

ISF_EXTENSIONS = [".json", ".json.xz", ".json.gz", ".json.bz2"]
"""List of accepted extensions for ISF files"""

SYMBOL_SERVER_URL = "http://msdl.microsoft.com/download/symbols"

if hasattr(sys, "frozen") and sys.frozen:
    # Ensure we include the executable's directory as the base for plugins and symbols
    PLUGINS_PATH = [
        os.path.abspath(os.path.join(os.path.dirname(sys.executable), "plugins"))
    ] + PLUGINS_PATH
    SYMBOL_BASEPATHS = [
        os.path.abspath(os.path.join(os.path.dirname(sys.executable), "symbols"))
    ] + SYMBOL_BASEPATHS

BANG = "!"
"""Constant used to delimit table names from type names when referring to a symbol"""

AUTOMAGIC_CONFIG_PATH = "automagic"
"""The root section within the context configuration for automagic values"""

LOGLEVEL_INFO = 20
"""Logging level for information data, showed when use the requests any logging: -v"""
LOGLEVEL_DEBUG = 10
"""Logging level for debugging data, showed when the user requests more logging detail: -vv"""
LOGLEVEL_V = 9
"""Logging level for the lowest "extra" level of logging: -vvv"""
LOGLEVEL_VV = 8
"""Logging level for two levels of detail: -vvvv"""
LOGLEVEL_VVV = 7
"""Logging level for three levels of detail: -vvvvv"""
LOGLEVEL_VVVV = 6
"""Logging level for four levels of detail: -vvvvvv"""


SQLITE_CACHE_PERIOD = "-3 days"
"""SQLite time modifier for how long each item is valid in the cache for"""

IDENTIFIERS_FILENAME = "identifier.cache"
"""Default location to record information about available identifiers"""

CACHE_SQLITE_SCHEMA_VERSION = 1
"""Version for the sqlite3 cache schema"""

BUG_URL = "https://github.com/volatilityfoundation/volatility3/issues"

ProgressCallback = Optional[Callable[[float, str], None]]
"""Type information for ProgressCallback objects"""

OS_CATEGORIES = ["windows", "mac", "linux"]


class Parallelism(enum.IntEnum):
    """An enumeration listing the different types of parallelism applied to
    volatility."""

    Off = 0
    Threading = 1
    Multiprocessing = 2


PARALLELISM = Parallelism.Off
"""Default value to the parallelism setting used throughout volatility"""

ISF_MINIMUM_SUPPORTED = (2, 0, 0)
"""The minimum supported version of the Intermediate Symbol Format"""
ISF_MINIMUM_DEPRECATED = (3, 9, 9)
"""The highest version of the ISF that's deprecated (usually higher than supported)"""
OFFLINE = False
"""Whether to go online to retrieve missing/necessary JSON files"""

REMOTE_ISF_URL = None  # 'http://localhost:8000/banners.json'
"""Remote URL to query for a list of ISF addresses"""

DOWNLOAD_TIMEOUT = 30
"""Length of time (in seconds) to wait for another process to download a resource before using it"""

###
# DEPRECATED VALUES
###

_deprecated_LINUX_BANNERS_FILENAME = os.path.join(CACHE_PATH, "linux_banners.cache")
"""This value is deprecated and is no longer used within volatility"""

_deprecated_MAC_BANNERS_PATH = os.path.join(CACHE_PATH, "mac_banners.cache")
"""This value is deprecated and is no longer used within volatility"""

_deprecated_IDENTIFIERS_PATH = os.path.join(CACHE_PATH, IDENTIFIERS_FILENAME)
"""This value is deprecated in favour of CACHE_PATH joined to IDENTIFIER_FILENAME"""


def __getattr__(name):
    deprecated_tag = "_deprecated_"
    if name in [
        x[len(deprecated_tag) :] for x in globals() if x.startswith(deprecated_tag)
    ]:
        warnings.warn(f"{name} is deprecated", FutureWarning)
        return globals()[f"{deprecated_tag}{name}"]

    return getattr(__import__(__name__), name)



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from volatility3.framework import interfaces, deprecation
from volatility3.plugins.windows.malware import malfind

vollog = logging.getLogger(__name__)


class Malfind(
    interfaces.plugins.PluginInterface,
    deprecation.PluginRenameClass,
    replacement_class=malfind.Malfind,
    removal_date="2026-06-07",
):
    """Lists process memory ranges that potentially contain injected code (deprecated)."""

    _required_framework_version = (2, 22, 0)
    _version = (1, 1, 0)



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Iterable, Tuple

from volatility3.framework import renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class Envars(plugins.PluginInterface):
    """Lists processes with their environment variables"""

    _required_framework_version = (2, 13, 0)
    _version = (2, 0, 1)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    @classmethod
    def get_task_env_variables(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
        env_area_max_size: int = 8192,
    ) -> Iterable[Tuple[str, str]]:
        """Yields environment variables for a given task.

        Args:
            context: The plugin's operational context.
            task: The task object from which to extract environment variables.
            env_area_max_size: Maximum allowable size for the environment variables area.
                Tasks exceeding this size will be skipped. Default is 8192.

        Yields:
            Tuples of (key, value) representing each environment variable.
        """

        task_name = utility.array_to_string(task.comm)
        task_pid = task.pid
        env_start = task.mm.env_start
        env_end = task.mm.env_end
        env_area_size = env_end - env_start
        if not (0 < env_area_size <= env_area_max_size):
            vollog.debug(
                f"Task {task_pid} {task_name} appears to have environment variables of size "
                f"{env_area_size} bytes which fails the sanity checking, will not extract "
                "any envars."
            )
            return None

        # Get process layer to read envars from
        proc_layer_name = task.add_process_layer()
        if proc_layer_name is None:
            return None
        proc_layer = context.layers[proc_layer_name]

        # Ensure the entire buffer is readable to prevent relying on exception handling
        if not proc_layer.is_valid(env_start, env_area_size):
            # Not mapped / swapped out
            vollog.debug(
                f"Unable to read environment variables for {task_pid} {task_name} starting at "
                f" virtual address 0x{env_start:x} for {env_area_size} bytes, will not "
                "extract any envars."
            )
            return None

        # Read the full task environment variable buffer.
        envar_data = proc_layer.read(env_start, env_area_size)

        # Parse envar data, envars are null terminated, keys and values are separated by '='
        envar_data = envar_data.rstrip(b"\x00")
        for envar_pair in envar_data.split(b"\x00"):
            try:
                env_key, env_value = envar_pair.decode(
                    encoding="utf8", errors="replace"
                ).split("=", 1)
            except ValueError:
                # Some legitimate programs, like 'avahi-daemon', avoid reallocating the args
                # and instead exploit the fact that the environment variables area is contiguous
                # to the args. This allows them to include a longer process name in the listing,
                # causing overwrites and incorrect results. In such cases, it's better to abort
                # the current task rather than displaying misleading or incorrect output.
                break

            yield env_key, env_value

    def _generator(self, tasks):
        """Generates a listing of processes along with environment variables"""

        # walk the process list and return the envars
        for task in tasks:
            if task.is_kernel_thread:
                continue

            task_pid = task.pid
            task_name = utility.array_to_string(task.comm)
            task_ppid = task.get_parent_pid()

            for env_key, env_value in self.get_task_env_variables(self.context, task):
                yield (0, (task_pid, task_ppid, task_name, env_key, env_value))

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        tasks = pslist.PsList.list_tasks(
            self.context, self.config["kernel"], filter_func=filter_func
        )

        headers = [
            ("PID", int),
            ("PPID", int),
            ("COMM", str),
            ("KEY", str),
            ("VALUE", str),
        ]

        return renderers.TreeGrid(headers, self._generator(tasks))



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import collections
import functools
import logging
import math
import struct
from typing import Any, Dict, Iterable, List, Optional, Tuple

from volatility3 import classproperty
from volatility3.framework import constants, exceptions, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import linear

vollog = logging.getLogger(__name__)

INTEL_TRANSLATION_DEBUGGING = False


class Intel(linear.LinearlyMappedLayer):
    """Translation Layer for the Intel IA32 memory mapping."""

    _PAGE_BIT_PRESENT = 0
    _PAGE_BIT_PSE = 7  # Page Size Extension: 4 MB (or 2MB) page
    _PAGE_BIT_PROTNONE = 8
    _PAGE_BIT_PAT_LARGE = 12  # 2MB or 1GB pages

    _PAGE_PRESENT = 1 << _PAGE_BIT_PRESENT
    _PAGE_PSE = 1 << _PAGE_BIT_PSE
    _PAGE_PROTNONE = 1 << _PAGE_BIT_PROTNONE
    _PAGE_PAT_LARGE = 1 << _PAGE_BIT_PAT_LARGE

    _entry_format = "<I"
    _page_size_in_bits = 12
    _bits_per_register = 32
    # NOTE: _maxphyaddr is MAXPHYADDR as defined in the Intel specs *NOT* the maximum physical address
    _maxphyaddr = 32
    _maxvirtaddr = _maxphyaddr
    _structure = [("page directory", 10, True), ("page table", 10, False)]
    _direct_metadata = collections.ChainMap(
        {"architecture": "Intel32"},
        {"mapped": True},
        interfaces.layers.TranslationLayerInterface._direct_metadata,
    )

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            context=context, config_path=config_path, name=name, metadata=metadata
        )
        self._base_layer = self.config["memory_layer"]
        self._swap_layers: List[str] = []
        self._page_map_offset = self.config["page_map_offset"]

        # Assign constants
        self._initial_position = min(self._maxvirtaddr, self._bits_per_register) - 1
        self._initial_entry = (
            self._mask(self._page_map_offset, self._initial_position, 0) | 0x1
        )
        self._entry_size = struct.calcsize(self._entry_format)
        self._entry_number = self.page_size // self._entry_size
        self._canonical_prefix = self._mask(
            (1 << self._bits_per_register) - 1,
            self._bits_per_register,
            self._maxvirtaddr,
        )

        # These can vary depending on the type of space
        self._index_shift = math.ceil(math.log2(struct.calcsize(self._entry_format)))

    @classproperty
    @functools.lru_cache
    def page_shift(cls) -> int:
        """Page shift for the intel memory layers."""
        return cls._page_size_in_bits

    @classproperty
    @functools.lru_cache
    def page_size(cls) -> int:
        """Page size for the intel memory layers.

        All Intel layers work on 4096 byte pages
        """
        return 1 << cls._page_size_in_bits

    @classproperty
    @functools.lru_cache
    def page_mask(cls) -> int:
        """Page mask for the intel memory layers."""
        return ~(cls.page_size - 1)

    @classproperty
    @functools.lru_cache
    def bits_per_register(cls) -> int:
        """Returns the bits_per_register to determine the range of an
        IntelTranslationLayer."""
        return cls._bits_per_register

    @classproperty
    @functools.lru_cache
    def minimum_address(cls) -> int:
        return 0

    @classproperty
    @functools.lru_cache
    def maximum_address(cls) -> int:
        return (1 << cls._maxvirtaddr) - 1

    @classproperty
    def structure(cls) -> List[Tuple[str, int, bool]]:
        return cls._structure

    @staticmethod
    def _mask(value: int, high_bit: int, low_bit: int) -> int:
        """Returns the bits of a value between highbit and lowbit inclusive."""
        high_mask = (1 << (high_bit + 1)) - 1
        low_mask = (1 << low_bit) - 1
        mask = high_mask ^ low_mask
        return value & mask

    @staticmethod
    def _page_is_valid(entry: int) -> bool:
        """Returns whether a particular page is valid based on its entry."""
        return bool(entry & 1)

    @staticmethod
    def _page_is_dirty(entry: int) -> bool:
        """Returns whether a particular page is dirty based on its entry."""
        return bool(entry & (1 << 6))

    def canonicalize(self, addr: int) -> int:
        """Canonicalizes an address by performing an appropriate sign extension on the higher addresses"""
        if self._bits_per_register <= self._maxvirtaddr:
            return addr & self.address_mask
        elif addr < (1 << self._maxvirtaddr - 1):
            return addr
        return self._mask(addr, self._maxvirtaddr, 0) + self._canonical_prefix

    def decanonicalize(self, addr: int) -> int:
        """Removes canonicalization to ensure an address fits within the correct range if it has been canonicalized

        This will produce an address outside the range if the canonicalization is incorrect
        """
        if addr < (1 << self._maxvirtaddr - 1):
            return addr
        return addr ^ self._canonical_prefix

    def _translate(self, offset: int) -> Tuple[int, int, str]:
        """Translates a specific offset based on paging tables.

        Returns the translated offset, the contiguous pagesize that the
        translated address lives in and the layer_name that the address
        lives in
        """
        entry, position = self._translate_entry(offset & self.page_mask)

        # Now we're done
        if not self._page_is_valid(entry):
            raise exceptions.PagedInvalidAddressException(
                self.name,
                offset,
                position + 1,
                entry,
                f"Page Fault at entry {hex(entry)} in page entry",
            )

        pfn = self._pte_pfn(entry)
        page_offset = self._mask(offset, position, 0)
        page = pfn << self.page_shift | page_offset

        return page, 1 << (position + 1), self._base_layer

    def _pte_pfn(self, entry: int) -> int:
        """Extracts the page frame number (PFN) from the page table entry (PTE) entry"""
        return self._mask(entry, self._maxphyaddr - 1, 0) >> self.page_shift

    @functools.lru_cache(maxsize=1024)
    def _translate_entry(self, page_address: int) -> int:
        """Translates a page address based on paging tables.

        Args:
            page_address: The page base address

        Returns:
            the translated entry value
        """
        # Setup the entry and how far we are through the offset
        # Position maintains the number of bits left to process
        # We or with 0x1 to ensure our page_map_offset is always valid
        position = self._initial_position
        entry = self._initial_entry

        if not (
            self.minimum_address
            <= (page_address & self.address_mask)
            <= self.maximum_address
        ):
            raise exceptions.PagedInvalidAddressException(
                self.name,
                page_address,
                position + 1,
                entry,
                "Entry outside virtual address range: " + hex(entry),
            )

        # Run through the offset in various chunks
        for name, size, large_page in self._structure:
            # Check we're valid
            if not self._page_is_valid(entry):
                raise exceptions.PagedInvalidAddressException(
                    self.name,
                    page_address,
                    position + 1,
                    entry,
                    "Page Fault at entry " + hex(entry) + " in table " + name,
                )

            # Grab the base address of the table we'll be getting the next entry from
            base_address = self._mask(
                entry, self._maxphyaddr - 1, size + self._index_shift
            )

            table = self._get_valid_table(base_address)
            if table is None:
                raise exceptions.PagedInvalidAddressException(
                    self.name,
                    page_address,
                    position + 1,
                    entry,
                    "Page Fault at entry " + hex(entry) + " in table " + name,
                )

            # Figure out how much of the offset we should be using
            start = position
            position -= size
            index = self._mask(page_address, start, position + 1) >> (position + 1)

            # Read the data for the next entry
            entry_data_start = index << self._index_shift
            entry_data = table[entry_data_start : entry_data_start + self._entry_size]

            if INTEL_TRANSLATION_DEBUGGING:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Entry {hex(entry)} at index {hex(index)} gives data {hex(struct.unpack(self._entry_format, entry_data)[0])} as {name}",
                )

            # Read out the new entry from memory
            (entry,) = struct.unpack(self._entry_format, entry_data)

            # Check if we're a large page
            if large_page and (entry & self._PAGE_PSE):
                # Mask off the PAT bit
                if entry & self._PAGE_PAT_LARGE:
                    entry -= self._PAGE_PAT_LARGE
                # We're a large page, the rest is finished below
                # If we want to implement PSE-36, it would need to be done here
                break

        return entry, position

    @functools.lru_cache(maxsize=1025)
    def _get_valid_table(self, base_address: int) -> Optional[bytes]:
        """Extracts the table, validates it and returns it if it's valid."""
        try:
            table = self._context.layers.read(
                self._base_layer, base_address, self.page_size
            )
        except exceptions.InvalidAddressException:
            return None

        ####
        # If the table is entirely duplicates, then mark the whole table as bad
        # This is because Windows 10 onwards has a tendency to map unused pages as present
        # This had the following consequences:
        #  - Used very litle physical memory
        #  - Exploded virtual memory
        #  - Causes *scan plugins to take multiple hours to complete even on small images

        # Previous versions of volatility would ignore a page during a scan when it matched
        # the one directly preceding it in physical memory.
        # This could trip if only two pages were identical and still required enumerating all
        # the invalid pages (which itself was quite time consuming)

        # For this reason, volatility 3 shifted to looking at entire page tables (1,024 pages)
        # and if all the pages mapped to the same place the table wouuld be skipped
        # This could also be applied to the Directory level as well as the Table level, allowing
        # Volatility to skip huge sections of virtual memory very efficiently, without missing
        # any pages that were distinct within a particular page table (or directory).

        # In order to work at this level, the logic was moved out of the scanning component and
        # directly into the layer logic itself.  This does have the side effect of preventing
        # entirely duplicated page tables from reporting as present, however, the trade off between
        # Windows 10+ reduced scanning times (common amongst scan plugins) versus incorrectly reporting
        # entire page tables of identically mapped repeating *valid* data (rare) was accepted in favour
        # of the more common occurance.
        if table == table[: self._entry_size] * self._entry_number:
            return None
        return table

    def is_valid(self, offset: int, length: int = 1) -> bool:
        """Returns whether the address offset can be translated to a valid
        address."""
        try:
            # TODO: Consider reimplementing this, since calls to mapping can call is_valid
            return all(
                self._context.layers[layer].is_valid(mapped_offset)
                for _, _, mapped_offset, _, layer in self.mapping(offset, length)
            )
        except exceptions.InvalidAddressException:
            return False

    def is_dirty(self, offset: int) -> bool:
        """Returns whether the page at offset is marked dirty"""
        return self._page_is_dirty(self._translate_entry(offset & self.page_mask)[0])

    def mapping(
        self, offset: int, length: int, ignore_errors: bool = False
    ) -> Iterable[Tuple[int, int, int, int, str]]:
        """Returns a sorted iterable of (offset, sublength, mapped_offset, mapped_length, layer)
        mappings.

        This allows translation layers to provide maps of contiguous
        regions in one layer
        """
        stashed_offset = stashed_mapped_offset = stashed_size = stashed_mapped_size = (
            stashed_map_layer
        ) = None
        for offset, size, mapped_offset, mapped_size, map_layer in self._mapping(
            offset, length, ignore_errors
        ):
            if (
                stashed_offset is None
                or (stashed_offset + stashed_size != offset)
                or (stashed_mapped_offset + stashed_mapped_size != mapped_offset)
                or (stashed_map_layer != map_layer)
            ):
                # The block isn't contiguous
                if stashed_offset is not None:
                    yield (
                        stashed_offset,
                        stashed_size,
                        stashed_mapped_offset,
                        stashed_mapped_size,
                        stashed_map_layer,
                    )
                # Update all the stashed values after output
                stashed_offset = offset
                stashed_mapped_offset = mapped_offset
                stashed_size = size
                stashed_mapped_size = mapped_size
                stashed_map_layer = map_layer
            else:
                # Part of an existing block
                stashed_size += size
                stashed_mapped_size += mapped_size
        # Yield whatever's left
        if (
            stashed_offset is not None
            and stashed_mapped_offset is not None
            and stashed_size is not None
            and stashed_mapped_size is not None
            and stashed_map_layer is not None
        ):
            yield (
                stashed_offset,
                stashed_size,
                stashed_mapped_offset,
                stashed_mapped_size,
                stashed_map_layer,
            )

    def _mapping(
        self, offset: int, length: int, ignore_errors: bool = False
    ) -> Iterable[Tuple[int, int, int, int, str]]:
        """Returns a sorted iterable of (offset, sublength, mapped_offset, mapped_length, layer)
        mappings.

        This allows translation layers to provide maps of contiguous
        regions in one layer
        """
        if length == 0:
            try:
                mapped_offset, _, layer_name = self._translate(offset)
                if not self._context.layers[layer_name].is_valid(mapped_offset):
                    raise exceptions.InvalidAddressException(
                        layer_name=layer_name, invalid_address=mapped_offset
                    )
            except exceptions.InvalidAddressException:
                if not ignore_errors:
                    raise
                return None
            yield offset, length, mapped_offset, length, layer_name
            return None
        while length > 0:
            skip_mask = None
            try:
                chunk_offset, page_size, layer_name = self._translate(offset)
                # Page align the chunk size value
                chunk_size = min(page_size - (offset % page_size), length)
                if not self._context.layers[layer_name].is_valid(
                    chunk_offset, chunk_size
                ):
                    # Virtual -> physical is contiguous in the chunk_size range.
                    # If we fail, we can jump directly to the end as we know all bytes in between
                    # aren't mapped (virtually and) physically anyway.
                    skip_mask = chunk_size - 1
                    raise exceptions.InvalidAddressException(
                        layer_name=layer_name, invalid_address=chunk_offset
                    )
            except (
                exceptions.PagedInvalidAddressException,
                exceptions.InvalidAddressException,
            ) as excp:
                if not ignore_errors:
                    raise
                if skip_mask is None:
                    # We can jump more if we know where the page fault occured
                    if isinstance(excp, exceptions.PagedInvalidAddressException):
                        skip_mask = (1 << excp.invalid_bits) - 1
                    else:
                        skip_mask = (1 << self._page_size_in_bits) - 1
                length_diff = skip_mask + 1 - (offset & skip_mask)
                length -= length_diff
                offset += length_diff
            else:
                yield offset, chunk_size, chunk_offset, chunk_size, layer_name
                length -= chunk_size
                offset += chunk_size

    @property
    def dependencies(self) -> List[str]:
        """Returns a list of the lower layer names that this layer is dependent
        upon."""
        return [self._base_layer] + self._swap_layers

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.TranslationLayerRequirement(
                name="memory_layer", optional=False
            ),
            requirements.LayerListRequirement(name="swap_layers", optional=True),
            requirements.IntRequirement(name="page_map_offset", optional=False),
            requirements.IntRequirement(name="kernel_virtual_offset", optional=True),
            requirements.StringRequirement(name="kernel_banner", optional=True),
        ]


class IntelPAE(Intel):
    """Class for handling Physical Address Extensions for Intel
    architectures."""

    _entry_format = "<Q"
    _bits_per_register = 32
    _maxphyaddr = 40
    _maxvirtaddr = 32
    _structure = [
        ("page directory pointer", 2, False),
        ("page directory", 9, True),
        ("page table", 9, False),
    ]
    _direct_metadata = collections.ChainMap({"pae": True}, Intel._direct_metadata)


class Intel32e(Intel):
    """Class for handling 64-bit (32-bit extensions) for Intel
    architectures."""

    _direct_metadata = collections.ChainMap(
        {"architecture": "Intel64"}, Intel._direct_metadata
    )
    _entry_format = "<Q"
    _bits_per_register = 64
    _maxphyaddr = 52
    _maxvirtaddr = 48
    _structure = [
        ("page map layer 4", 9, False),
        ("page directory pointer", 9, True),
        ("page directory", 9, True),
        ("page table", 9, False),
    ]


class WindowsMixin(Intel):
    @staticmethod
    def _page_is_valid(entry: int) -> bool:
        """Returns whether a particular page is valid based on its entry.

        Windows uses additional "available" bits to store flags
        These flags allow windows to determine whether a page is still valid

        Bit 11 is the transition flag, and Bit 10 is the prototype flag

        For more information, see Windows Internals (6th Ed, Part 2, pages 268-269)
        """
        return bool((entry & 1) or ((entry & 1 << 11) and not entry & 1 << 10))

    def _translate_swap(
        self, layer: Intel, offset: int, bit_offset: int
    ) -> Tuple[int, int, str]:
        try:
            return super()._translate(offset)
        except exceptions.PagedInvalidAddressException as excp:
            entry = excp.entry
            tbit = bool(entry & (1 << 11))
            pbit = bool(entry & (1 << 10))
            unknown_bit = bool(entry & (1 << 7))
            n = (entry >> 1) & 0xF
            vbit = bool(entry & 1)
            if (not tbit and not pbit and not vbit and unknown_bit) and (
                (entry >> bit_offset) != 0
            ):
                swap_offset = entry >> bit_offset << excp.invalid_bits

                if layer.config.get("swap_layers", False):
                    swap_layer_name = layer.config.get(
                        interfaces.configuration.path_join(
                            "swap_layers", "swap_layers" + str(n)
                        ),
                        None,
                    )
                    if swap_layer_name:
                        return swap_offset, 1 << excp.invalid_bits, swap_layer_name
                raise exceptions.SwappedInvalidAddressException(
                    layer_name=excp.layer_name,
                    invalid_address=excp.invalid_address,
                    invalid_bits=excp.invalid_bits,
                    entry=excp.entry,
                    swap_offset=swap_offset,
                )
            raise


### These must be full separate classes so that JSON configs re-create them properly


class WindowsIntel(WindowsMixin, Intel):
    def _translate(self, offset):
        return self._translate_swap(self, offset, self._page_size_in_bits)


class WindowsIntelPAE(WindowsMixin, IntelPAE):
    def _translate(self, offset: int) -> Tuple[int, int, str]:
        return self._translate_swap(self, offset, self._bits_per_register)


class WindowsIntel32e(WindowsMixin, Intel32e):
    # TODO: Fix appropriately in a future release.
    # Currently just a temporary workaround to deal with custom bit flag
    # in the PFN field for pages in transition state.
    # See https://github.com/volatilityfoundation/volatility3/pull/475
    _maxphyaddr = 45

    def _translate(self, offset: int) -> Tuple[int, int, str]:
        return self._translate_swap(self, offset, self._bits_per_register // 2)


class LinuxMixin(Intel):
    @functools.cached_property
    def _register_mask(self) -> int:
        return (1 << self._bits_per_register) - 1

    @functools.cached_property
    def _physical_mask(self) -> int:
        # From kernels 4.18 the physical mask is dynamic: See AMD SME, Intel Multi-Key Total
        # Memory Encryption and CONFIG_DYNAMIC_PHYSICAL_MASK: 94d49eb30e854c84d1319095b5dd0405a7da9362
        physical_mask = (1 << self._maxphyaddr) - 1
        # TODO: Come back once SME support is available in the framework
        return physical_mask

    @functools.cached_property
    def page_mask(self) -> int:
        # Note that within the Intel class it's a class method. However, since it uses
        # complement operations and we are working in Python, it would be more careful to
        # limit it to the architecture's pointer size.
        return ~(self.page_size - 1) & self._register_mask

    @functools.cached_property
    def _physical_page_mask(self) -> int:
        return self.page_mask & self._physical_mask

    @functools.cached_property
    def _pte_pfn_mask(self) -> int:
        return self._physical_page_mask

    @functools.cached_property
    def _pte_flags_mask(self) -> int:
        return ~self._pte_pfn_mask & self._register_mask

    def _pte_flags(self, pte) -> int:
        return pte & self._pte_flags_mask

    def _is_pte_present(self, entry: int) -> bool:
        return (
            self._pte_flags(entry) & (self._PAGE_PRESENT | self._PAGE_PROTNONE)
        ) != 0

    def _page_is_valid(self, entry: int) -> bool:
        # Overrides the Intel static method with the Linux-specific implementation
        return self._is_pte_present(entry)

    def _pte_needs_invert(self, entry) -> bool:
        # Entries that were set to PROT_NONE (PAGE_PRESENT) are inverted
        # A clear PTE shouldn't be inverted. See f19f5c4
        return entry and not (entry & self._PAGE_PRESENT)

    def _protnone_mask(self, entry: int) -> int:
        """Gets a mask to XOR with the page table entry to get the correct PFN"""
        return self._register_mask if self._pte_needs_invert(entry) else 0

    def _pte_pfn(self, entry: int) -> int:
        """Extracts the page frame number from the page table entry"""
        pfn = entry ^ self._protnone_mask(entry)
        return (pfn & self._pte_pfn_mask) >> self.page_shift


class LinuxIntel(LinuxMixin, Intel):
    pass


class LinuxIntelPAE(LinuxMixin, IntelPAE):
    pass


class LinuxIntel32e(LinuxMixin, Intel32e):
    # In the Linux kernel, the __PHYSICAL_MASK_SHIFT is a mask used to extract the
    # physical address from a PTE. In Volatility3, this is referred to as _maxphyaddr.
    #
    # Until kernel version 4.17, Linux x86-64 used a 46-bit mask. With commit
    # b83ce5ee91471d19c403ff91227204fb37c95fb2, this was extended to 52 bits,
    # applying to both 4 and 5-level page tables.
    #
    # We initially used 52 bits for all Intel 64-bit systems, but this produced incorrect
    # results for PROT_NONE pages. Since the mask value is defined by a preprocessor macro,
    # it's difficult to detect the exact bit shift used in the current kernel.
    # Using 46 bits has proven reliable for our use case, as seen in tools like crashtool.
    _maxphyaddr = 46



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import datetime
from typing import List, Tuple, Iterable


from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.plugins import timeliner
from volatility3.plugins.linux import pslist


class Boottime(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Shows the time the system was started"""

    _required_framework_version = (2, 11, 0)
    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
        ]

    @classmethod
    def get_time_namespaces_bootime(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> Iterable[Tuple[int, int, int, str, datetime.datetime]]:
        """Enumerates tasks' boot times based on their time namespaces.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
            pids: Pid list
            unique: Filter unique time namespaces

        Yields:
            A tuple with the fields to show in the plugin output.
        """
        time_namespace_ids = set()
        for task in pslist.PsList.list_tasks(context, vmlinux_module_name):
            time_namespace_id = task.get_time_namespace_id()
            # If it cannot get the time namespace i.e. kernels < 5.6, this still works
            # using None to just get the first tasks
            if time_namespace_id in time_namespace_ids:
                continue
            time_namespace_ids.add(time_namespace_id)
            boottime = task.get_boottime(root_time_namespace=False)

            fields = (
                time_namespace_id,
                boottime,
            )
            yield fields

    def _generator(self):
        for (
            time_namespace_id,
            boottime,
        ) in self.get_time_namespaces_bootime(
            self.context,
            self.config["kernel"],
        ):
            fields = [
                time_namespace_id or renderers.NotAvailableValue(),
                boottime,
            ]
            yield 0, fields

    def generate_timeline(self):
        for (
            time_namespace_id,
            boottime,
        ) in self.get_time_namespaces_bootime(
            self.context,
            self.config["kernel"],
        ):
            description = f"System boot time for time namespace {time_namespace_id}"

            yield description, timeliner.TimeLinerType.CREATED, boottime

    def run(self):
        columns = [
            ("TIME NS", int),
            ("Boot Time", datetime.datetime),
        ]
        return renderers.TreeGrid(columns, self._generator())
